# -*- coding: utf-8 -*- #
# Copyright 2023 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command for adding layer7 ddos defense threshold config to security policies."""


from googlecloudsdk.api_lib.compute import base_classes
from googlecloudsdk.api_lib.compute.security_policies import client
from googlecloudsdk.calliope import arg_parsers
from googlecloudsdk.calliope import base
from googlecloudsdk.calliope import exceptions
from googlecloudsdk.command_lib.compute.security_policies import flags
from googlecloudsdk.command_lib.compute.security_policies import security_policies_utils


class AddLayer7DdosDefenseThresholdConfigHelper(object):
  r"""Add a layer7 ddos defense threshold config to a Compute Engine security policy.

  *{command}* is used to add layer7 ddos defense threshold configs to security policies.

  ## EXAMPLES

  To add a layer7 ddos defense threshold config run the following command:

    $ {command} NAME \
       --threshold-config-name=my-threshold-config-name \
       --auto-deploy-load-threshold=0.7 \
       --auto-deploy-confidence-threshold=0.8 \
       --auto-deploy-impacted-baseline-threshold=0.1 \
       --auto-deploy-expiration-sec=4800
  """

  SECURITY_POLICY_ARG = None

  @classmethod
  def Args(cls, parser, support_granularity_config):
    """Adds the arguments for the command."""
    cls.SECURITY_POLICY_ARG = flags.SecurityPolicyArgument()
    cls.SECURITY_POLICY_ARG.AddArgument(parser, operation_type='update')
    parser.add_argument(
        '--threshold-config-name',
        required=True,
        help='The name for the threshold config.',
    )
    parser.add_argument(
        '--auto-deploy-load-threshold',
        type=float,
        required=False,
        help=(
            "The threshold on backend's load, over which auto-deploy takes"
            ' action.'
        ),
    )
    parser.add_argument(
        '--auto-deploy-confidence-threshold',
        type=float,
        required=False,
        help=(
            'The threshold of the confidence of an identified attack, over'
            ' which auto-deploy takes action.'
        ),
    )
    parser.add_argument(
        '--auto-deploy-impacted-baseline-threshold',
        type=float,
        required=False,
        help=(
            'The threshold on the estimated impact to the baseline traffic of a'
            ' suggested mitigation, below which auto-deploy takes action.'
        ),
    )
    parser.add_argument(
        '--auto-deploy-expiration-sec',
        type=int,
        required=False,
        help='The duration of actions, if any, taken by auto-deploy.',
    )
    if support_granularity_config:
      parser.add_argument(
          '--detection-load-threshold',
          type=float,
          required=False,
          help=(
              "The threshold on backend's load, over which adaptive protection"
              ' detects an attack.'
          ),
      )
      parser.add_argument(
          '--detection-absolute-qps',
          type=float,
          required=False,
          help=(
              'The absolute QPS of the incoming traffic, over which adaptive'
              ' protection detects an attack.'
          ),
      )
      parser.add_argument(
          '--detection-relative-to-baseline-qps',
          type=float,
          required=False,
          help=(
              'The QPS of the incoming traffic relative to the average baseline'
              ' QPS, over which adaptive protection detects an attack.'
          ),
      )
      parser.add_argument(
          '--traffic-granularity-configs',
          type=arg_parsers.ArgList(
              max_length=2,
              custom_delim_char=';',
              element_type=arg_parsers.ArgDict(
                  min_length=2,
                  max_length=3,
                  spec={
                      'type': str,
                      'value': str,
                      'enableEachUniqueValue': bool,
                  },
                  required_keys=['type'],
              ),
          ),
          required=False,
          metavar='type=TYPE[,value=VALUE][,enableEachUniqueValue=ENABLE_EACH_UNIQUE_VALUE];...',
          help=(
              'Specify up to 2 configs matching a specifc type/value of'
              ' traffic.'
          ),
      )

  @classmethod
  def Run(cls, release_track, args, support_granularity_config):
    """Runs the command."""
    holder = base_classes.ComputeApiHolder(release_track)
    ref = cls.SECURITY_POLICY_ARG.ResolveAsResource(args, holder.resources)
    security_policy = client.SecurityPolicy(
        ref=ref, compute_client=holder.client
    )
    existing_security_policy = security_policy.Describe()[0]

    adaptive_protection_config = (
        existing_security_policy.adaptiveProtectionConfig
    )
    if (
        adaptive_protection_config is None
        or adaptive_protection_config.layer7DdosDefenseConfig is None
        or not adaptive_protection_config.layer7DdosDefenseConfig.enable
    ):
      raise exceptions.InvalidArgumentException(
          '--threshold-config-name',
          'Must enable adaptive protection layer 7 ddos defense before adding a'
          ' threshold config',
      )

    threshold_config = (
        security_policies_utils.CreateLayer7DdosDefenseThresholdConfig(
            holder.client,
            args,
            support_granularity_config=support_granularity_config,
        )
    )
    adaptive_protection_config.layer7DdosDefenseConfig.thresholdConfigs.append(
        threshold_config
    )

    updated_security_policy = holder.client.messages.SecurityPolicy(
        adaptiveProtectionConfig=adaptive_protection_config,
        fingerprint=existing_security_policy.fingerprint,
    )

    return security_policy.Patch(security_policy=updated_security_policy)


@base.UniverseCompatible
@base.ReleaseTracks(
    base.ReleaseTrack.ALPHA,
    base.ReleaseTrack.BETA,
    base.ReleaseTrack.GA,
    base.ReleaseTrack.PREVIEW,
)
class AddLayer7DdosDefenseThresholdConfigGA(base.UpdateCommand):
  r"""Add a layer7 ddos defense threshold config to a Compute Engine security policy.

  *{command}* is used to add layer7 ddos defense threshold configs to security policies.

  ## EXAMPLES

  To add a layer7 ddos defense threshold config run the following command:

    $ {command} NAME \
       --threshold-config-name=my-threshold-config-name \
       --auto-deploy-load-threshold=0.7 \
       --auto-deploy-confidence-threshold=0.8 \
       --auto-deploy-impacted-baseline-threshold=0.1 \
       --auto-deploy-expiration-sec=4800
       --detection-load-threshold=0.4
       --detection-absolute-qps=1000
       --detection-relative-to-baseline-qps=2.0
       --traffic-granularity-configs=type=HTTP_HEADER_HOST,value=www.my-test-host.com;type=HTTP_PATH,enableEachUniqueValue=true
  """

  _support_granularity_config = True

  @classmethod
  def Args(cls, parser):
    AddLayer7DdosDefenseThresholdConfigHelper.Args(
        parser, support_granularity_config=cls._support_granularity_config
    )

  def Run(self, args):
    return AddLayer7DdosDefenseThresholdConfigHelper.Run(
        self.ReleaseTrack(),
        args,
        support_granularity_config=self._support_granularity_config,
    )



# -*- coding: utf-8 -*- #
# Copyright 2022 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command for invalidating an access approval request."""


import textwrap

from googlecloudsdk.api_lib.access_approval import requests
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.access_approval import request_name


class Invalidate(base.Command):
  """Invalidate an Access Approval request.

  Invalidate an Access Approval request. This will raise an error if the request
  does not exist or is not in an approved state.
  """

  detailed_help = {
      'EXAMPLES':
          textwrap.dedent("""\
        To invalidate an approval request using its name (e.g. projects/12345/approvalRequests/abc123), run:

          $ {command} projects/12345/approvalRequests/abc123
        """),
  }

  @staticmethod
  def Args(parser):
    """Add command-specific args."""
    request_name.Args(parser)

  def Run(self, args):
    """This is what gets called when the user runs this command.

    Args:
      args: an argparse namespace. All the arguments that were provided to this
        command invocation.

    Returns:
      Some value that we want to have printed later.
    """
    return requests.Invalidate(request_name.GetName(args))



# -*- coding: utf-8 -*- #
# Copyright 2020 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command for access approval list requests."""


import textwrap

from googlecloudsdk.api_lib.access_approval import requests
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.access_approval import parent


class List(base.ListCommand):
  """List Access Approval requests.

  List Access Approval requests by parent (project/folder/organization).
  """

  detailed_help = {
      'EXAMPLES':
          textwrap.dedent("""\
        To list all approval requests owned by project my-project-123, run:

          $ {command} --project=my-project-123 --state=all

        To list pending approval requests owned by organization 999, run:

          $ {command} --organization=999

        or

          $ {command} --organization=999 --state=pending

        Note that the user needs to have permission
        accessapproval.requests.list on the project/folder/organization
        """),
  }

  @staticmethod
  def Args(parser):
    """Add command-specific args."""
    parent.Args(parser)
    parser.add_argument(
        '--state',
        default='pending',
        help='filter for request state')
    base.URI_FLAG.RemoveFromParser(parser)

  def Run(self, args):
    """This is what gets called when the user runs this command.

    Args:
      args: an argparse namespace. All the arguments that were provided to this
        command invocation.

    Returns:
      Some value that we want to have printed later.
    """
    p = parent.GetParent(args)
    return requests.List(parent=p, filter=(
        args.state.upper() if args.state else None))



# -*- coding: utf-8 -*- #
# Copyright 2015 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""`gcloud app versions describe` command."""


from googlecloudsdk.api_lib.app import appengine_api_client
from googlecloudsdk.calliope import base


class Describe(base.DescribeCommand):
  """Display all data about an existing version."""

  @staticmethod
  def Args(parser):
    parser.add_argument(
        '--service', '-s', required=True,
        help='The service corresponding to the version to show.')
    parser.add_argument('version', help='The ID of the version to show.')

  def Run(self, args):
    api_client = appengine_api_client.GetApiClientForTrack(self.ReleaseTrack())
    return api_client.GetVersionResource(service=args.service,
                                         version=args.version)



# -*- coding: utf-8 -*- #
# Copyright 2020 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command line processing utilities for cloud access bindings."""


import re

from apitools.base.py import encoding
from googlecloudsdk.api_lib.accesscontextmanager import util
from googlecloudsdk.calliope import exceptions as calliope_exceptions
from googlecloudsdk.command_lib.accesscontextmanager import common
from googlecloudsdk.core import exceptions as core_exceptions
from googlecloudsdk.core import properties
from googlecloudsdk.core import resources
from googlecloudsdk.core.util import iso_duration
from googlecloudsdk.core.util import times

_WEB_APP_NAMES_14_DAY_SESSION_LENGTH = frozenset(['Cloud Console'])

def AddUpdateMask(ref, args, req):
  """Hook to add update mask."""
  del ref
  update_mask = []
  if args.IsKnownAndSpecified('level'):
    update_mask.append('access_levels')
  if args.IsKnownAndSpecified('dry_run_level'):
    update_mask.append('dry_run_access_levels')
  if args.IsKnownAndSpecified('session_length'):
    update_mask.append('session_settings')
  if args.IsKnownAndSpecified('binding_file'):
    update_mask.append('scoped_access_settings')

  if not update_mask:
    raise calliope_exceptions.MinimumArgumentException(
        ['--level', '--dry_run_level', '--session-length', '--binding-file']
    )

  req.updateMask = ','.join(update_mask)
  return req


def AddUpdateMaskAlpha(ref, args, req):
  """Hook to add update mask in Alpha track."""
  del ref
  update_mask = []
  if args.IsKnownAndSpecified('level'):
    update_mask.append('access_levels')
  if args.IsKnownAndSpecified('dry_run_level'):
    update_mask.append('dry_run_access_levels')
  if args.IsKnownAndSpecified(
      'restricted_client_application_client_ids'
  ) or args.IsKnownAndSpecified('restricted_client_application_names'):
    update_mask.append('restricted_client_applications')
  if args.IsKnownAndSpecified('session_length'):
    update_mask.append('session_settings')
  if args.IsKnownAndSpecified('binding_file'):
    update_mask.append('scoped_access_settings')

  if not update_mask:
    raise calliope_exceptions.MinimumArgumentException([
        '--level',
        '--dry_run_level',
        '--restricted_client_application_names',
        '--restricted_client_application_client_ids',
        '--session-length',
        '--binding-file',
    ])

  req.updateMask = ','.join(update_mask)
  return req


def ProcessOrganization(ref, args, req):
  """Hook to process organization input."""
  del ref, args
  if req.parent is not None:
    return req

  org = properties.VALUES.access_context_manager.organization.Get()
  if org is None:
    raise calliope_exceptions.RequiredArgumentException(
        '--organization',
        'The attribute can be set in the following ways: \n'
        + '- provide the argument `--organization` on the command line \n'
        + '- set the property `access_context_manager/organization`',
    )

  org_ref = resources.REGISTRY.Parse(
      org, collection='accesscontextmanager.organizations'
  )
  req.parent = org_ref.RelativeName()
  return req


def ProcessRestrictedClientApplicationsAlpha(unused_ref, args, req):
  """Hook to process restricted client applications input in Alpha track."""
  del unused_ref
  return _ProcessRestrictedClientApplications(args, req, version='v1alpha')


def _ProcessRestrictedClientApplications(args, req, version=None):
  """Process restricted client applications input for the given version."""
  # Processing application client ids if available
  if args.IsKnownAndSpecified('restricted_client_application_client_ids'):
    client_ids = args.restricted_client_application_client_ids
    restricted_client_application_refs = (
        _MakeRestrictedClientApplicationsFromIdentifiers(
            client_ids,
            'restricted_client_application_client_ids',
            version=version,
        )
    )
    # req.gcpUserAccessBinding is None when no access levels are specified
    # during update. Access Levels are optional when updating restricted client
    # applications, but they are required when creating a new binding.
    if req.gcpUserAccessBinding is None:
      req.gcpUserAccessBinding = util.GetMessages(
          version=version
      ).GcpUserAccessBinding()
    for restricted_client_application_ref in restricted_client_application_refs:
      req.gcpUserAccessBinding.restrictedClientApplications.append(
          restricted_client_application_ref
      )
  # processing application names if available
  if args.IsKnownAndSpecified('restricted_client_application_names'):
    client_names = args.restricted_client_application_names
    restricted_client_application_refs = (
        _MakeRestrictedClientApplicationsFromIdentifiers(
            client_names,
            'restricted_client_application_names',
            version=version,
        )
    )
    # req.gcpUserAccessBinding is None when no access levels are specified
    # during update. Access Levels are optional when updating restricted client
    # applications, but they are required when creating a new binding.
    if req.gcpUserAccessBinding is None:
      req.gcpUserAccessBinding = util.GetMessages(
          version=version
      ).GcpUserAccessBinding()
    for restricted_client_application_ref in restricted_client_application_refs:
      req.gcpUserAccessBinding.restrictedClientApplications.append(
          restricted_client_application_ref
      )
  return req


def _MakeRestrictedClientApplicationsFromIdentifiers(
    app_identifiers, arg_name, version=None
):
  """Parse restricted client applications and return their resource references."""
  resource_refs = []
  if app_identifiers is not None:
    app_identifiers = [
        # remove empty strings
        identifier
        for identifier in app_identifiers
        if identifier
    ]
    for app_identifier in app_identifiers:
      if arg_name == 'restricted_client_application_client_ids':
        try:
          resource_refs.append(
              util.GetMessages(version=version).Application(
                  clientId=app_identifier
              )
          )
        except:
          raise calliope_exceptions.InvalidArgumentException(
              '--{}'.format('restricted_client_application_client_ids'),
              'Unable to parse input. The input must be of type string[].',
          )
      elif arg_name == 'restricted_client_application_names':
        try:
          resource_refs.append(
              util.GetMessages(version=version).Application(name=app_identifier)
          )
        except:
          raise calliope_exceptions.InvalidArgumentException(
              '--{}'.format('restricted_client_application_names'),
              'Unable to parse input. The input must be of type string[].',
          )
      else:
        raise calliope_exceptions.InvalidArgumentException(
            '--{}'.format('arg_name'),
            'The input is not valid for Restricted Client Applications.',
        )
  return resource_refs


def _ParseLevelRefs(req, param, is_dry_run):
  """Parse level strings and return their resource references."""
  level_inputs = req.gcpUserAccessBinding.accessLevels
  if is_dry_run:
    level_inputs = req.gcpUserAccessBinding.dryRunAccessLevels

  level_refs = []
  level_inputs = [level_input for level_input in level_inputs if level_input]
  if not level_inputs:
    return level_refs

  arg_name = '--dry_run_level' if is_dry_run else '--level'

  for level_input in level_inputs:
    try:
      level_ref = resources.REGISTRY.Parse(
          level_input,
          params=param,
          collection='accesscontextmanager.accessPolicies.accessLevels',
      )
    except:
      raise calliope_exceptions.InvalidArgumentException(
          '--{}'.format(arg_name),
          'The input must be the full identifier for the access level, '
          'such as `accessPolicies/123/accessLevels/abc`.',
      )
    level_refs.append(level_ref)
  return level_refs


def ProcessLevels(ref, args, req):
  """Hook to format levels and validate all policies."""
  del ref  # Unused
  policies_to_check = {}

  param = {}
  policy_ref = None
  if args.IsKnownAndSpecified('policy'):
    try:
      policy_ref = resources.REGISTRY.Parse(
          args.GetValue('policy'),
          collection='accesscontextmanager.accessPolicies',
      )
    except:
      raise calliope_exceptions.InvalidArgumentException(
          '--policy',
          'The input must be the full identifier for the access policy, '
          'such as `123` or `accessPolicies/123.',
      )
    param = {'accessPoliciesId': policy_ref.Name()}
    policies_to_check['--policy'] = policy_ref.RelativeName()
  else:
    del policy_ref

  # Parse level and dry run level
  level_refs = (
      _ParseLevelRefs(req, param, is_dry_run=False)
      if args.IsKnownAndSpecified('level')
      else []
  )
  dry_run_level_refs = (
      _ParseLevelRefs(req, param, is_dry_run=True)
      if args.IsKnownAndSpecified('dry_run_level')
      else []
  )

  # Validate all refs in each level ref belong to the same policy
  level_parents = [x.Parent() for x in level_refs]
  dry_run_level_parents = [x.Parent() for x in dry_run_level_refs]
  if not all(x == level_parents[0] for x in level_parents):
    raise ConflictPolicyException(['--level'])
  if not all(x == dry_run_level_parents[0] for x in dry_run_level_parents):
    raise ConflictPolicyException(['--dry-run-level'])

  # Validate policies of level, dry run level and policy inputs are the same
  if level_parents:
    policies_to_check['--level'] = level_parents[0].RelativeName()
  if dry_run_level_parents:
    policies_to_check['--dry-run-level'] = dry_run_level_parents[
        0
    ].RelativeName()
  flags_to_complain = list(policies_to_check.keys())
  flags_to_complain.sort()  # Sort for test purpose.
  policies_values = list(policies_to_check.values())
  if not all(x == policies_values[0] for x in policies_values):
    raise ConflictPolicyException(flags_to_complain)

  # Set formatted level fields in the request
  if level_refs:
    req.gcpUserAccessBinding.accessLevels = [
        x.RelativeName() for x in level_refs
    ]
  if dry_run_level_refs:
    req.gcpUserAccessBinding.dryRunAccessLevels = [
        x.RelativeName() for x in dry_run_level_refs
    ]
  return req


def ProcessSessionLength(string):
  """Process the session-length argument into an acceptable form for GCSL session settings."""

  # If we receive the empty string then return a negative duration. This will
  # signal to the request processor that sessionSettings should be cleared.
  # This is primarily used for clearing bindings on calls to update, and is a
  # no-op for calls to create.

  duration = (
      times.ParseDuration(string) if string else iso_duration.Duration(hours=-1)
  )

  # TODO(b/346781832)
  if duration.total_seconds > iso_duration.Duration(days=1).total_seconds:
    raise calliope_exceptions.InvalidArgumentException(
        '--session-length',
        'The session length cannot be greater than one day.',
    )
  # Format for Google protobuf Duration
  return '{}s'.format(int(duration.total_seconds))


def ProcessSessionSettings(unused_ref, args, req):
  """Hook to process GCSL session settings.

    When --session-length=0 make sure the sessionLengthEnabled is set to false.

    Throw an error if --session-reauth-method or --use-oidc-max-age are set
    without --session-length.

  Args:
      unused_ref: Unused
      args: The command line arguments
      req: The request object

  Returns:
    The modified request object.

  Raises:
    calliope_exceptions.InvalidArgumentException: If arguments are incorrectly
    set.
  """
  del unused_ref
  if args.IsKnownAndSpecified('session_length'):
    if args.IsKnownAndSpecified(
        'restricted_client_application_client_ids'
    ) or args.IsKnownAndSpecified('restricted_client_application_names'):
      raise calliope_exceptions.InvalidArgumentException(
          '--session-length',
          'Cannot set session length on restricted client applications. Use '
          'scoped access settings.',
      )
    session_length = times.ParseDuration(
        req.gcpUserAccessBinding.sessionSettings.sessionLength
    ).total_seconds
    if session_length < 0:  # Case where --session_length=''
      req.gcpUserAccessBinding.sessionSettings = None
    elif session_length == 0:  # Case where we disable session
      req.gcpUserAccessBinding.sessionSettings.sessionLengthEnabled = False
    else:  # Normal case
      req.gcpUserAccessBinding.sessionSettings.sessionLengthEnabled = True
  else:
    if args.IsKnownAndSpecified('session_reauth_method'):
      raise calliope_exceptions.InvalidArgumentException(
          '--session_reauth_method',
          'Cannot set --session_reauth_method without --session-length',
      )
    # Clear all default session settings from the request if --session-length is
    # unspecified
    req.gcpUserAccessBinding.sessionSettings = None

  return req


def _CamelCase2SnakeCase(name):
  s1 = re.compile('([a-z0-9])([A-Z])').sub(r'\1_\2', name)
  return re.sub('_[A-Z]+', lambda m: m.group(0).lower(), s1)


def ProcessFilter(unused_ref, args, req):
  """Hook to process filter. Covert camel case to snake case."""
  del unused_ref
  if args.IsKnownAndSpecified('filter'):
    # Only pass filter to handler if it contains principal
    if 'principal' in args.filter:
      filter_str = _CamelCase2SnakeCase(args.filter)
      req.filter = filter_str
  return req


class ConflictPolicyException(core_exceptions.Error):
  """For conflict policies from inputs."""

  def __init__(self, parameter_names):
    super(ConflictPolicyException, self).__init__(
        'Invalid value for [{0}]: Ensure that the {0} resources are '
        'all from the same policy.'.format(
            ', '.join(['{0}'.format(p) for p in parameter_names])
        )
    )


def _TryGetAccessLevelResources(
    param, access_levels, field_name, error_message
):
  """Try to get the access level cloud resources that correspond to the `access levels`.

  Args:
    param: The parameters to pass to the resource registry
    access_levels: The access levels to turn into cloud resources
    field_name: The name of the field to use in the error message
    error_message: The error message to use if the access levels cannot be
      parsed

  Returns:
    The access level cloud resources that correspond to the `access levels`.
  """
  access_level_resources = []
  access_level_inputs = [
      access_level for access_level in access_levels if access_level
  ]

  for access_level_input in access_level_inputs:
    try:
      access_level_resources.append(
          resources.REGISTRY.Parse(
              access_level_input,
              params=param,
              collection='accesscontextmanager.accessPolicies.accessLevels',
          )
      )
    except:
      raise calliope_exceptions.InvalidArgumentException(
          '--{}'.format(field_name),
          error_message,
      )

  return access_level_resources


def _TryGetPolicyCloudResource(policy, field_name, error_message):
  """Try to get the policy cloud resource that corresponds to the `policy`.

  Args:
    policy: The policy to turn into a cloud resource
    field_name: The name of the field to use in the error message
    error_message: The error message to use if the policy cannot be parsed

  Returns:
    The policy cloud resource that corresponds to the `policy`.
  """
  try:
    return resources.REGISTRY.Parse(
        policy,
        collection='accesscontextmanager.accessPolicies',
    )
  except:
    raise calliope_exceptions.InvalidArgumentException(
        '--{}'.format(field_name), error_message
    )


def _ProcessScopesInScopedAccessSettings(req):
  """Validates the scope in the scoped access settings."""

  def _ValidateScopeInScopedAccessSettingsUniqueness(scoped_access_settings):
    scopes = [str(x.scope) for x in scoped_access_settings]
    if len(scopes) != len(set(scopes)):
      raise calliope_exceptions.InvalidArgumentException(
          '--binding-file',
          'ScopedAccessSettings in the binding-file must be unique.',
      )

  def _IsClientScopeSet(client_scope):
    if not client_scope:
      return False
    if client_scope.restrictedClientApplication:
      restricted_client_application_dict = encoding.MessageToDict(
          client_scope.restrictedClientApplication
      )
      if not restricted_client_application_dict:
        return False
      # Check for None or empty string
      for key in restricted_client_application_dict.keys():
        if not restricted_client_application_dict[key]:
          return False
      return True
    elif (
        hasattr(client_scope, 'restrictedProject')
        and client_scope.restrictedProject
    ):
      restricted_project_dict = encoding.MessageToDict(
          client_scope.restrictedProject
      )
      if not restricted_project_dict:
        return False
      for key in restricted_project_dict.keys():
        if not restricted_project_dict[key]:
          return False
      return True
    return False

  def _ValidateScopeInScopedAccessSettingIsNotEmpty(scoped_access_setting):
    if not scoped_access_setting.scope or not _IsClientScopeSet(
        scoped_access_setting.scope.clientScope
    ):
      raise calliope_exceptions.InvalidArgumentException(
          '--binding-file',
          'ScopedAccessSettings in the binding-file must have a scope.',
      )

  def _Start(req):
    scoped_access_settings = req.gcpUserAccessBinding.scopedAccessSettings
    _ValidateScopeInScopedAccessSettingsUniqueness(scoped_access_settings)
    for scoped_access_setting in scoped_access_settings:
      _ValidateScopeInScopedAccessSettingIsNotEmpty(scoped_access_setting)

  _Start(req)


def _ProcessAccessSettingsInScopedAccessSettings(req):
  """Validates the access settings in the scoped access settings."""

  def _IsAccessSettingsSet(access_settings):
    if not access_settings:
      return False
    access_settings_dict = encoding.MessageToDict(access_settings)
    if not access_settings_dict:
      return False
    # Check for None or empty arrays
    for key in access_settings_dict.keys():
      if not access_settings_dict[key]:
        return False
    return True

  def _ValidateAccessSettingsInScopedAccessSettingAtLeastOneIsNotEmpty(
      access_settings, dry_run_settings
  ):
    if not _IsAccessSettingsSet(access_settings) and not _IsAccessSettingsSet(
        dry_run_settings
    ):
      raise calliope_exceptions.InvalidArgumentException(
          '--binding-file',
          'ScopedAccessSettings in the binding-file must have at least one of'
          ' activeSettings or dryRunSettings set.',
      )

  def _Start(req):
    scoped_access_settings = req.gcpUserAccessBinding.scopedAccessSettings
    for scoped_access_setting in scoped_access_settings:
      _ValidateAccessSettingsInScopedAccessSettingAtLeastOneIsNotEmpty(
          scoped_access_setting.activeSettings,
          scoped_access_setting.dryRunSettings,
      )

  _Start(req)


def _ProcessAccessLevelsInScopedAccessSettings(args, req):
  """Process the access levels in the scoped access settings."""

  def _ValidateBelongsToSamePolicy(
      access_level_resources,
      dry_run_access_level_resources,
      policy_resource,
      parameter_names,
  ):
    """Validate that the access levels and policy belong to the same policy."""
    combined_access_level = (
        access_level_resources + dry_run_access_level_resources
    )
    if combined_access_level:
      # Check that all access levels are from the same policy
      access_level_resources_parents = [
          x.Parent() for x in combined_access_level
      ]
      if not all(
          x == access_level_resources_parents[0]
          for x in access_level_resources_parents
      ):
        raise ConflictPolicyException(parameter_names)

      # Check that the policy is the same as the access levels
      if (
          policy_resource
          and access_level_resources_parents
          and (
              policy_resource.RelativeName()
              != access_level_resources_parents[0].RelativeName()
          )
      ):
        raise ConflictPolicyException(['--policy'] + parameter_names)

  def _ReplaceAccessLevelsInAccessSettingsWithRelativeNames(
      access_settings, access_level_resources
  ):
    """Replace the access levels in the scoped access settings with relative names.

    For example,

    {
      'activeSettings': {
        'accessLevels': [
          'accessPolicies/123/accessLevels/access_level_1'
        ]
      }
    }

    is replaced with:

    {
      'activeSettings': {
        'accessLevels': [
          access_level_resources.RelativeName()
        ]
      }
    }

    Args:
      access_settings: The access settings to replace the access levels in.
      access_level_resources: The access level resources to replace the access
        levels with.
    """
    # Set the relative names of the access levels in the request
    if access_level_resources:
      access_settings.accessLevels = [
          x.RelativeName() for x in access_level_resources
      ]

  def _GetAccessLevelResources(policy_resource, access_levels):
    """Get the access level resources from the scoped access settings.

    Args:
      policy_resource: The policy resource
      access_levels: The access levels to turn into cloud resources. For
        example, ['accessPolicies/123/accessLevels/access_level_1']

    Returns:
      The access level cloud resources that correspond to the `access levels`.
      For example,
      ['https://accesscontextmanager.googleapis.com/v1/accessPolicies/123/accessLevels/access_level_1']
    """
    param = (
        {}
        if not policy_resource
        else {'accessPoliciesId': policy_resource.Name()}
    )
    # Obtain the access level resources
    access_level_resources = []
    if access_levels:
      access_level_resources = _TryGetAccessLevelResources(
          param,
          access_levels,
          'binding-file',
          'Access levels in ScopedAccessSettings must contain the full'
          ' identifier. For example:'
          ' `accessPolicies/123/accessLevels/access_level_1',
      )
    return access_level_resources

  def _Start(args, req):
    policy_resource = None
    if args.IsKnownAndSpecified('policy'):
      # Obtain the policy resource
      policy_resource = _TryGetPolicyCloudResource(
          args.GetValue('policy'),
          'policy',
          'The input must be the full identifier for the access policy, '
          'such as `123` or `accessPolicies/123.',
      )

    scoped_access_settings = req.gcpUserAccessBinding.scopedAccessSettings
    access_level_resources_sample = []
    dry_run_access_level_resources_sample = []
    for scoped_access_setting in scoped_access_settings:
      # Obtain the access level resources
      access_level_resources = []
      if (
          scoped_access_setting.activeSettings
          and scoped_access_setting.activeSettings.accessLevels
      ):
        access_level_resources = _GetAccessLevelResources(
            policy_resource, scoped_access_setting.activeSettings.accessLevels
        )
        access_level_resources_sample.append(access_level_resources[0])

      # Obtain the dry run access level resources
      dry_run_access_level_resources = []
      if (
          scoped_access_setting.dryRunSettings
          and scoped_access_setting.dryRunSettings.accessLevels
      ):
        dry_run_access_level_resources = _GetAccessLevelResources(
            policy_resource,
            scoped_access_setting.dryRunSettings.accessLevels,
        )
        dry_run_access_level_resources_sample.append(
            dry_run_access_level_resources[0]
        )
      _ValidateBelongsToSamePolicy(
          access_level_resources,
          dry_run_access_level_resources,
          policy_resource,
          ['--binding-file'],
      )
      _ReplaceAccessLevelsInAccessSettingsWithRelativeNames(
          scoped_access_setting.activeSettings, access_level_resources
      )
      _ReplaceAccessLevelsInAccessSettingsWithRelativeNames(
          scoped_access_setting.dryRunSettings, dry_run_access_level_resources
      )

    # Validate that all access levels in all scoped access settings belong to
    # the same policy
    _ValidateBelongsToSamePolicy(
        access_level_resources_sample,
        dry_run_access_level_resources_sample,
        policy_resource,
        ['--binding-file'],
    )

    # Obtain the global access level resource for the first access level defined
    # in the request
    global_access_level_resources = []
    if req.gcpUserAccessBinding.accessLevels:
      try:
        global_access_level_resources = _GetAccessLevelResources(
            policy_resource, req.gcpUserAccessBinding.accessLevels
        )
      except calliope_exceptions.InvalidArgumentException:
        # Ignore error because global access levels will be processed later
        pass
    if not global_access_level_resources:
      try:
        global_access_level_resources = _GetAccessLevelResources(
            policy_resource, req.gcpUserAccessBinding.dryRunAccessLevels
        )
      except calliope_exceptions.InvalidArgumentException:
        # Ignore error because global access levels will be processed later
        pass

    # Validated that scoped and global access levels belong to the same policy
    _ValidateBelongsToSamePolicy(
        access_level_resources_sample,
        global_access_level_resources,
        policy_resource,
        ['--binding-file', '--level', '--dry-run-level'],
    )

  _Start(args, req)


def _ProcessSessionSettingsInScopedAccessSettings(req):
  """Process the session settings in the scoped access settings."""

  is_update_request = hasattr(req, 'name')
  # For update requests, we don't know if a binding is for a federated
  # principal without making a GET request first. To avoid this, we assume it
  # could be federated and allow session lengths > 1d, and rely on backend
  # validation to enforce 1d max for non-federated principals.
  has_federated_principal = is_update_request or (
      hasattr(req.gcpUserAccessBinding, 'principal')
      and req.gcpUserAccessBinding.principal
      and req.gcpUserAccessBinding.principal.federatedPrincipal
  )

  def _ValidateSessionSettings(session_settings, scope):
    if session_settings is None:
      return
    if session_settings.sessionLength is None:
      raise calliope_exceptions.InvalidArgumentException(
          '--binding-file',
          'SessionSettings within ScopedAccessSettings must include a session'
          'length.',
      )
    session_length_sec = times.ParseDuration(
        session_settings.sessionLength
    ).total_seconds
    max_session_length = iso_duration.Duration(days=1)
    # TODO(b/486106966): Investigate if we can do a better check for v1alpha.
    if (
        properties.VALUES.access_context_manager.enable_gcsl.GetBool()
        and scope
        and scope.clientScope
    ):
      if has_federated_principal:
        if (
            scope.clientScope.restrictedClientApplication
            and scope.clientScope.restrictedClientApplication.name
            in _WEB_APP_NAMES_14_DAY_SESSION_LENGTH
        ):
          max_session_length = iso_duration.Duration(days=14)
        else:
          max_session_length = iso_duration.Duration(days=90)

    if session_length_sec > max_session_length.total_seconds:
      raise calliope_exceptions.InvalidArgumentException(
          '--binding-file',
          'SessionLength within ScopedAccessSettings must not be greater than'
          ' {} second(s)'.format(max_session_length.total_seconds),
      )
    if session_length_sec < 3600 and session_length_sec != 0:
      raise calliope_exceptions.InvalidArgumentException(
          '--binding-file',
          'SessionLength within ScopedAccessSettings must not be less than '
          '1 hour, unless it is zero',
      )

  def _InferEmptySessionSettingsFields(session_settings):
    # When sessionReauthMethod is absent, infer LOGIN
    if session_settings.sessionReauthMethod is None:
      v1_messages = util.GetMessages('v1')
      if isinstance(session_settings, v1_messages.SessionSettings):
        session_settings.sessionReauthMethod = (
            v1_messages.SessionSettings.SessionReauthMethodValueValuesEnum.LOGIN
        )
      else:
        session_settings.sessionReauthMethod = util.GetMessages(
            'v1alpha'
        ).SessionSettings.SessionReauthMethodValueValuesEnum.LOGIN
    # When sessionLengthEnabled is absent, infer True if SessionLength is
    # greater than zero, otherwise infer false.
    if session_settings.sessionLengthEnabled is None:
      session_length = times.ParseDuration(
          session_settings.sessionLength
      ).total_seconds
      if session_length > 0:
        session_settings.sessionLengthEnabled = True
      else:
        session_settings.sessionLengthEnabled = False
    # When useOidcMaxAge is absent, infer False
    if session_settings.useOidcMaxAge is None:
      session_settings.useOidcMaxAge = False

  def _Start(req):
    scoped_access_settings = req.gcpUserAccessBinding.scopedAccessSettings
    for s in scoped_access_settings:
      if not s.activeSettings:
        continue
      session_settings = s.activeSettings.sessionSettings
      if not session_settings:
        continue
      _ValidateSessionSettings(session_settings, s.scope)
      _InferEmptySessionSettingsFields(session_settings)

  _Start(req)


def _ValidateRestrictedProjectScope(args, req):
  """Validate restricted project scope."""
  if (
      not req.gcpUserAccessBinding
      or not req.gcpUserAccessBinding.scopedAccessSettings
  ):
    return

  has_restricted_project = False
  for sas in req.gcpUserAccessBinding.scopedAccessSettings:
    if (
        sas.scope
        and sas.scope.clientScope
        and getattr(sas.scope.clientScope, 'restrictedProject', None)
    ):
      has_restricted_project = True
      # If restrictedProject is present:
      # group_key must not be used.
      if hasattr(args, 'group_key') and args.IsKnownAndSpecified('group_key'):
        raise calliope_exceptions.InvalidArgumentException(
            '--group-key',
            'Restricted project scope cannot be used with --group-key; use'
            ' --federated-principal instead.',
        )
      # active_settings.session_settings must be present.
      if not sas.activeSettings or not sas.activeSettings.sessionSettings:
        raise calliope_exceptions.InvalidArgumentException(
            '--binding-file',
            'Restricted project scope must have active session settings set.',
        )
      # access_levels and dry_run_access_levels must be empty
      if sas.activeSettings and sas.activeSettings.accessLevels:
        raise calliope_exceptions.InvalidArgumentException(
            '--binding-file',
            'Restricted project scope cannot be used with access levels in'
            ' activeSettings.',
        )
      if sas.dryRunSettings and sas.dryRunSettings.accessLevels:
        raise calliope_exceptions.InvalidArgumentException(
            '--binding-file',
            'Restricted project scope cannot be used with access levels in'
            ' dryRunSettings.',
        )
  if has_restricted_project:
    # federated_principal must be used with restricted project scope.
    if properties.VALUES.access_context_manager.enable_gcsl.GetBool():
      if hasattr(args, 'federated_principal') and not args.IsKnownAndSpecified(
          'federated_principal'
      ):
        raise calliope_exceptions.InvalidArgumentException(
            '--binding-file',
            'When using a restricted project scope, --federated-principal must'
            ' be specified.',
        )


def ProcessScopedAccessSettings(unused_ref, args, req):
  """Hook to process and validate scoped access settings from the request."""

  def _ValidateRestrictedClientApplicationNamesAndClientIdsAreNotSpecified(
      args,
  ):
    legacy_prca_fields_specified = args.IsKnownAndSpecified(
        'restricted_client_application_names'
    ) or args.IsKnownAndSpecified('restricted_client_application_client_ids')
    if legacy_prca_fields_specified:
      raise calliope_exceptions.InvalidArgumentException(
          '--binding-file',
          'The binding-file cannot be specified at the same time as'
          ' `--restricted-client-application-names` or'
          ' `--restricted-client-application-client-ids`.',
      )

  def _Start(unused_ref, args, req):
    del unused_ref
    if not args.IsKnownAndSpecified('binding_file'):
      return req

    _ValidateRestrictedClientApplicationNamesAndClientIdsAreNotSpecified(args)
    _ProcessScopesInScopedAccessSettings(req)
    _ValidateRestrictedProjectScope(args, req)
    _ProcessAccessSettingsInScopedAccessSettings(req)
    _ProcessAccessLevelsInScopedAccessSettings(args, req)
    _ProcessSessionSettingsInScopedAccessSettings(req)

    return req

  return _Start(unused_ref, args, req)


class InvalidFormatError(common.ParseFileError):

  def __init__(self, path, reason):
    super(InvalidFormatError, self).__init__(
        path,
        (
            'Invalid format: {}\n\n'
            ' A binding-file is a YAML-formatted file'
            ' containing a single gcpUserAccessBinding.'
            ' For example:\n\n'
            '  scopedAccessSettings:\n'
            '  - scope:\n'
            '      clientScope:\n'
            '        restrictedClientApplication:\n'
            '          name: Cloud Console\n'
            '    activeSettings:\n'
            '      accessLevels:\n'
            '      - accessPolicies/123/accessLevels/access_level_1\n'
            '    dryRunSettings:\n'
            '      accessLevels:\n'
            '      - accessPolicies/123/accessLevels/dry_run_access_level_1\n'
            '  - scope:\n'
            '      clientScope:\n'
            '        restrictedClientApplication:\n'
            '          clientId: my_client_id.google.com\n'
            '    activeSettings:\n'
            '      accessLevels:\n'
            '      - accessPolicies/123/accessLevels/access_level_2\n'
            '    dryRunSetting:\n'
            '      accessLevels:\n'
            '      - accessPolicies/123/accessLevels/dry_run_access_level_2\n'
        ).format(
            reason,
        ),
    )


def ParseGcpUserAccessBindingFromBindingFile(api_version):
  """Parse a GcpUserAccessBinding from a YAML file.

  Args:
    api_version: str, the API version to use for parsing the messages

  Returns:
    A function that parses a GcpUserAccessBinding from a file.
  """

  def _ValidateSingleGcpUserAccessBinding(bindings):
    if len(bindings) > 1:
      raise calliope_exceptions.InvalidArgumentException(
          '--input-file',
          'The input file contains more than one GcpUserAccessBinding. '
          'Please specify only one GcpUserAccessBinding in the input file.',
      )

  def _ParseVersionedGcpUserAccessBindingFromBindingFile(path):
    bindings = common.ParseAccessContextManagerMessagesFromYaml(
        path, util.GetMessages(version=api_version).GcpUserAccessBinding, False
    )
    _ValidateSingleGcpUserAccessBinding(bindings)
    GcpUserAccessBindingStructureValidator(
        path, bindings[0], api_version
    ).Validate()
    return bindings[0]

  return _ParseVersionedGcpUserAccessBindingFromBindingFile


class GcpUserAccessBindingStructureValidator:
  """Validates a GcpUserAccessBinding structure against unrecognized fields."""

  def __init__(self, path, gcp_user_access_binding, api_version):
    self.path = path
    self.gcp_user_access_binding = gcp_user_access_binding
    self.api_version = api_version

  def Validate(self):
    """Validates the GcpUserAccessBinding structure."""
    self._ValidateAllFieldsRecognizedForGcpUserAccessBinding(
        self.gcp_user_access_binding
    )
    self._ValidateScopedAccessSettings(
        self.gcp_user_access_binding.scopedAccessSettings
    )

  def _ValidateScopedAccessSettings(self, scoped_access_settings_list):
    """Validates the ScopedAccessSettings structure."""
    if scoped_access_settings_list:
      for i in range(len(scoped_access_settings_list)):
        scoped_access_settings = scoped_access_settings_list[i]
        self._ValidateAllFieldsRecognized(scoped_access_settings)
        self._ValidateAccessScope(scoped_access_settings.scope)
        self._ValidateAccessSettings(scoped_access_settings.activeSettings)
        self._ValidateAccessSettings(scoped_access_settings.dryRunSettings)

  def _ValidateAccessScope(self, access_scope):
    """Validates the AccessScope structure."""
    if access_scope:
      self._ValidateAllFieldsRecognized(access_scope)
      self._ValidateClientScope(access_scope.clientScope)

  def _ValidateClientScope(self, client_scope):
    """Validates the AccessScopeType structure."""
    if client_scope:
      self._ValidateAllFieldsRecognized(client_scope)
      self._ValidateRestrictedClientApplication(
          client_scope.restrictedClientApplication
      )
      if (
          properties.VALUES.access_context_manager.enable_gcsl.GetBool()
          and hasattr(client_scope, 'restrictedProject')
      ):
        self._ValidateProject(client_scope.restrictedProject)

  def _ValidateRestrictedClientApplication(self, restricted_client_application):
    """Validates the RestrictedClientApplications."""
    if restricted_client_application:
      self._ValidateAllFieldsRecognized(restricted_client_application)

  def _ValidateProject(self, restricted_project):
    """Validates the Project."""
    if restricted_project:
      self._ValidateAllFieldsRecognized(restricted_project)

  def _ValidateSessionSettings(self, session_settings):
    """Validate the SessionSettings."""
    if session_settings:
      self._ValidateAllFieldsRecognized(session_settings)

  def _ValidateAccessSettings(self, access_settings):
    """Validates the AccessSettings structure."""
    if access_settings:
      self._ValidateAllFieldsRecognized(access_settings)
      self._ValidateSessionSettings(access_settings.sessionSettings)

  def _ValidateAllFieldsRecognizedForGcpUserAccessBinding(
      self, gcp_user_access_binding
  ):
    """Validates that all fields in the GcpUserAccessBinding are recognized.

    Note:Because ScopedAccessSettings is the only field supported in the
    GcpUserAccessBinding, a custom validation is required.

    Args:
      gcp_user_access_binding: The GcpUserAccessBinding to validate

    Raises:
      InvalidFormatError: if the GcpUserAccessBinding contains unrecognized
      fields
    """
    valid_fields = ['scopedAccessSettings']
    unrecognized_fields = set()
    empty_list = []
    if gcp_user_access_binding.accessLevels != empty_list:
      unrecognized_fields.add('accessLevels')
    if gcp_user_access_binding.dryRunAccessLevels != empty_list:
      unrecognized_fields.add('dryRunAccessLevels')
    if gcp_user_access_binding.groupKey is not None:
      unrecognized_fields.add('groupKey')
    if gcp_user_access_binding.name:
      unrecognized_fields.add('name')
    if (
        hasattr(gcp_user_access_binding, 'principal')
        and gcp_user_access_binding.principal is not None
    ):
      unrecognized_fields.add('principal')
    if gcp_user_access_binding.sessionSettings is not None:
      unrecognized_fields.add('sessionSettings')
    if gcp_user_access_binding.restrictedClientApplications:
      unrecognized_fields.add('restrictedClientApplications')
    if gcp_user_access_binding.all_unrecognized_fields():
      unrecognized_fields.update(
          gcp_user_access_binding.all_unrecognized_fields()
      )
    if unrecognized_fields:
      raise InvalidFormatError(
          self.path,
          '"{}" contains unrecognized fields: [{}]. Valid fields are: [{}].'
          .format(
              type(self.gcp_user_access_binding).__name__,
              ', '.join(unrecognized_fields),
              ', '.join(valid_fields),
          ),
      )

  def _ValidateAllFieldsRecognized(self, message):
    """Validates that all fields in the message are recognized.

    Args:
      message: object to validate

    Raises:
      InvalidFormatError: if the message contains unrecognized fields
    """
    unrecognized_fields_set = set(message.all_unrecognized_fields())
    message_type = type(message)
    valid_fields_list = [f.name for f in message_type.all_fields()]
    if message_type.__name__ == 'ClientScope':
      if not properties.VALUES.access_context_manager.enable_gcsl.GetBool():
        if 'restrictedProject' in valid_fields_list:
          valid_fields_list.remove('restrictedProject')
        if hasattr(message, 'restrictedProject') and message.restrictedProject:
          unrecognized_fields_set.add('restrictedProject')

    if unrecognized_fields_set:
      raise InvalidFormatError(
          self.path,
          '"{}" contains unrecognized fields: [{}]. Valid fields are: [{}]'
          .format(
              message_type.__name__,
              ', '.join(sorted(unrecognized_fields_set)),
              ', '.join(sorted(valid_fields_list)),
          ),
      )



# -*- coding: utf-8 -*- #
# Copyright 2026 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Run constants."""

# Common field names in Cloud Run (K8s-style) resources.
METADATA = 'metadata'
NAMESPACE = 'namespace'
SPEC = 'spec'
TEMPLATE = 'template'
ANNOTATIONS = 'annotations'
STATUS = 'status'
TRAFFIC = 'traffic'
PERCENT = 'percent'
SERVICE_ACCOUNT_NAME = 'serviceAccountName'
REVISION_NAME = 'revisionName'

# Execution environment values.
GEN1 = 'gen1'
GEN2 = 'gen2'

# SSH specific constants.
SSH_ROOT_USER = 'root'
SSH_HOST_KEY_ALIAS = 'cloud-run-default'
SSH_URL_TEMPLATE = 'wss://{region}.ssh.run.app/v4'
SSH_CA_PUBLIC_KEY_URL_TEMPLATE = (
    'https://www.gstatic.com/cloud-run/ssh-ca-public-keys/keys-{region}.pub'
)
SSH_CA_PUBLIC_KEY_URL_QUAL_TEMPLATE = (
    'https://www.gstatic.com/cloud-run/ssh-ca-public-keys/qual-keys-{region}.pub'
)



# -*- coding: utf-8 -*- #
# Copyright 2020 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Import target HTTP Proxies command."""


from apitools.base.py import exceptions as apitools_exceptions
from googlecloudsdk.api_lib.compute import base_classes
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.compute import exceptions as compute_exceptions
from googlecloudsdk.command_lib.compute import flags as compute_flags
from googlecloudsdk.command_lib.compute import operation_utils
from googlecloudsdk.command_lib.compute import scope as compute_scope
from googlecloudsdk.command_lib.compute.target_http_proxies import flags
from googlecloudsdk.command_lib.compute.target_http_proxies import target_http_proxies_utils
from googlecloudsdk.command_lib.export import util as export_util
from googlecloudsdk.core import yaml_validator
from googlecloudsdk.core.console import console_io


def _DetailedHelp():
  return {
      'brief':
          'Import a target HTTP proxy.',
      'DESCRIPTION':
          """\
          Imports a target HTTP proxy's configuration from a file.
          """,
      'EXAMPLES':
          """\
          A target HTTP proxy can be imported by running:

            $ {command} NAME --source=<path-to-file>
          """
  }


def _GetApiVersion(release_track):
  """Returns the API version based on the release track."""
  if release_track == base.ReleaseTrack.ALPHA:
    return 'alpha'
  elif release_track == base.ReleaseTrack.BETA:
    return 'beta'
  return 'v1'


def _GetSchemaPath(release_track, for_help=False):
  """Returns the resource schema path."""
  return export_util.GetSchemaPath(
      'compute',
      _GetApiVersion(release_track),
      'TargetHttpProxy',
      for_help=for_help)


def _SendInsertRequest(client, resources, target_http_proxy_ref,
                       target_http_proxy):
  """Sends Target HTTP Proxy insert request."""
  if target_http_proxies_utils.IsRegionalTargetHttpProxiesRef(
      target_http_proxy_ref):
    service = client.apitools_client.regionTargetHttpProxies
    operation = service.Insert(
        client.messages.ComputeRegionTargetHttpProxiesInsertRequest(
            project=target_http_proxy_ref.project,
            region=target_http_proxy_ref.region,
            targetHttpProxy=target_http_proxy))
  else:
    service = client.apitools_client.targetHttpProxies
    operation = service.Insert(
        client.messages.ComputeTargetHttpProxiesInsertRequest(
            project=target_http_proxy_ref.project,
            targetHttpProxy=target_http_proxy))

  return _WaitForOperation(resources, service, operation, target_http_proxy_ref,
                           'Inserting TargetHttpProxy')


def _SendPatchRequest(client, resources, target_http_proxy_ref,
                      target_http_proxy):
  """Make target HTTP proxy patch request."""
  if target_http_proxies_utils.IsRegionalTargetHttpProxiesRef(
      target_http_proxy_ref):
    console_message = ('Target HTTP Proxy [{0}] cannot be updated'.format(
        target_http_proxy_ref.Name()))
    raise NotImplementedError(console_message)

  service = client.apitools_client.targetHttpProxies
  operation = service.Patch(
      client.messages.ComputeTargetHttpProxiesPatchRequest(
          project=target_http_proxy_ref.project,
          targetHttpProxy=target_http_proxy_ref.Name(),
          targetHttpProxyResource=target_http_proxy))

  return _WaitForOperation(resources, service, operation, target_http_proxy_ref,
                           'Updating TargetHttpProxy')


def _WaitForOperation(resources, service, operation, target_http_proxy_ref,
                      message):
  """Waits for the TargetHttpProxy operation to finish."""
  if target_http_proxies_utils.IsRegionalTargetHttpProxiesRef(
      target_http_proxy_ref):
    collection = operation_utils.GetRegionalOperationsCollection()
  else:
    collection = operation_utils.GetGlobalOperationsCollection()

  return operation_utils.WaitForOperation(resources, service, operation,
                                          collection, target_http_proxy_ref,
                                          message)


def _Run(args, holder, target_http_proxy_arg, release_track):
  """Issues requests necessary to import target HTTP proxies."""
  client = holder.client
  resources = holder.resources

  target_http_proxy_ref = target_http_proxy_arg.ResolveAsResource(
      args,
      holder.resources,
      default_scope=compute_scope.ScopeEnum.GLOBAL,
      scope_lister=compute_flags.GetDefaultScopeLister(client))

  data = console_io.ReadFromFileOrStdin(args.source or '-', binary=False)

  try:
    target_http_proxy = export_util.Import(
        message_type=client.messages.TargetHttpProxy,
        stream=data,
        schema_path=_GetSchemaPath(release_track))
  except yaml_validator.ValidationError as e:
    raise compute_exceptions.ValidationError(str(e))

  # Get existing target HTTP proxy.
  try:
    target_http_proxy_old = target_http_proxies_utils.SendGetRequest(
        client, target_http_proxy_ref)
  except apitools_exceptions.HttpError as error:
    if error.status_code != 404:
      raise error
    # Target HTTP proxy does not exist, create a new one.
    return _SendInsertRequest(client, resources, target_http_proxy_ref,
                              target_http_proxy)

  if target_http_proxy_old == target_http_proxy:
    return

  console_io.PromptContinue(
      message=('Target Http Proxy [{0}] will be overwritten.').format(
          target_http_proxy_ref.Name()),
      cancel_on_no=True)

  # Populate id and fingerprint fields. These two fields are manually
  # removed from the schema files.
  target_http_proxy.id = target_http_proxy_old.id
  target_http_proxy.fingerprint = target_http_proxy_old.fingerprint

  # Unspecified fields are assumed to be cleared.
  cleared_fields = []
  if target_http_proxy.description is None:
    cleared_fields.append('description')

  # The REST API will reject requests without the UrlMap. However, we want to
  # avoid doing partial validations in the client and rely on server side
  # behavior.
  if target_http_proxy.urlMap is None:
    cleared_fields.append('urlMap')
  if release_track != base.ReleaseTrack.GA:
    if target_http_proxy.proxyBind is None:
      cleared_fields.append('proxyBind')

  with client.apitools_client.IncludeFields(cleared_fields):
    return _SendPatchRequest(client, resources, target_http_proxy_ref,
                             target_http_proxy)


@base.ReleaseTracks(
    base.ReleaseTrack.GA,
    base.ReleaseTrack.BETA,
    base.ReleaseTrack.ALPHA,
    base.ReleaseTrack.PREVIEW,
)
class Import(base.UpdateCommand):
  """Import a target HTTP Proxy."""

  detailed_help = _DetailedHelp()
  TARGET_HTTP_PROXY_ARG = None

  @classmethod
  def Args(cls, parser):
    cls.TARGET_HTTP_PROXY_ARG = flags.TargetHttpProxyArgument()
    cls.TARGET_HTTP_PROXY_ARG.AddArgument(parser, operation_type='import')
    export_util.AddImportFlags(
        parser, _GetSchemaPath(cls.ReleaseTrack(), for_help=True))

  def Run(self, args):
    holder = base_classes.ComputeApiHolder(self.ReleaseTrack())
    return _Run(args, holder, self.TARGET_HTTP_PROXY_ARG, self.ReleaseTrack())



# -*- coding: utf-8 -*- #
# Copyright 2020 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command for describing target gRPC proxies."""


from googlecloudsdk.api_lib.compute import base_classes
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.compute import scope as compute_scope
from googlecloudsdk.command_lib.compute.target_grpc_proxies import flags


def _DetailedHelp():
  return {
      'brief':
          'Display detailed information about a target gRPC proxy.',
      'DESCRIPTION':
          """\
        *{command}* displays all data associated with a Compute Engine
        target gRPC proxy.
      """,
      'EXAMPLES':
          """\
      To describe a global target gRPC proxy, run:

        $ {command} PROXY_NAME
      """,
  }


def _Run(holder, target_grpc_proxy_ref):
  """Issues requests necessary to describe Target gRPC Proxies."""
  client = holder.client
  request = client.messages.ComputeTargetGrpcProxiesGetRequest(
      **target_grpc_proxy_ref.AsDict())
  collection = client.apitools_client.targetGrpcProxies

  return client.MakeRequests([(collection, 'Get', request)])[0]


class Describe(base.DescribeCommand):
  """Display detailed information about a target gRPC proxy."""

  TARGET_GRPC_PROXY_ARG = None
  detailed_help = _DetailedHelp()

  @classmethod
  def Args(cls, parser):
    cls.TARGET_GRPC_PROXY_ARG = flags.TargetGrpcProxyArgument()
    cls.TARGET_GRPC_PROXY_ARG.AddArgument(parser, operation_type='describe')

  def Run(self, args):
    holder = base_classes.ComputeApiHolder(self.ReleaseTrack())
    target_grpc_proxy_ref = self.TARGET_GRPC_PROXY_ARG.ResolveAsResource(
        args, holder.resources, default_scope=compute_scope.ScopeEnum.GLOBAL)
    return _Run(holder, target_grpc_proxy_ref)



# -*- coding: utf-8 -*- #
# Copyright 2021 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command for deleting organization firewall policies."""


from googlecloudsdk.api_lib.compute import base_classes
from googlecloudsdk.api_lib.compute.firewall_policies import client
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.compute.firewall_policies import firewall_policies_utils
from googlecloudsdk.command_lib.compute.firewall_policies import flags
import six


@base.DefaultUniverseOnly
class Delete(base.DeleteCommand):
  """Delete a Compute Engine organization firewall policy.

  *{command}* is used to delete organization firewall policies. An organization
  firewall policy is a set of rules that controls access to various resources.
  """

  FIREWALL_POLICY_ARG = None

  @classmethod
  def Args(cls, parser):
    cls.FIREWALL_POLICY_ARG = flags.FirewallPolicyArgument(
        required=True, operation='delete')
    cls.FIREWALL_POLICY_ARG.AddArgument(parser, operation_type='delete')
    parser.add_argument(
        '--organization',
        help=('Organization in which the organization firewall policy is to be'
              ' deleted. Must be set if FIREWALL_POLICY is the short name.'))
    parser.display_info.AddCacheUpdater(flags.FirewallPoliciesCompleter)

  def Run(self, args):
    holder = base_classes.ComputeApiHolder(self.ReleaseTrack())
    ref = self.FIREWALL_POLICY_ARG.ResolveAsResource(
        args, holder.resources, with_project=False)
    org_firewall_policy = client.OrgFirewallPolicy(
        ref=ref,
        compute_client=holder.client,
        resources=holder.resources,
        version=six.text_type(self.ReleaseTrack()).lower())
    fp_id = firewall_policies_utils.GetFirewallPolicyId(
        org_firewall_policy, ref.Name(), organization=args.organization)
    return org_firewall_policy.Delete(fp_id=fp_id, only_generate_request=False)


Delete.detailed_help = {
    'EXAMPLES':
        """\
    To delete an organization firewall policy with ID ``123456789", run:

      $ {command} 123456789
    """,
    'IAM PERMISSIONS': """\
    To delete a firewall policy, the user must have the following
    permission: *`compute.firewallPolicies.delete`.

    To find predefined roles that contain those permissions, see the [Compute
    Engine IAM roles](https://cloud.google.com/compute/docs/access/iam).
    """
}



# -*- coding: utf-8 -*- #
# Copyright 2019 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ai-platform jobs cancel command."""


from googlecloudsdk.api_lib.ml_engine import jobs
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.ml_engine import flags
from googlecloudsdk.command_lib.ml_engine import jobs_util


DETAILED_HELP = {
    'EXAMPLES':
        """\
          To cancel a running AI Platform job named ``my-job'', run:

            $ {command} my-job
        """
}


def _AddCancelArgs(parser):
  flags.JOB_NAME.AddToParser(parser)


class Cancel(base.SilentCommand):
  """Cancel a running AI Platform job."""

  detailed_help = DETAILED_HELP

  @staticmethod
  def Args(parser):
    _AddCancelArgs(parser)

  def Run(self, args):
    return jobs_util.Cancel(jobs.JobsClient(), args.job)


_DETAILED_HELP = {
    'DESCRIPTION':
        """\
*{command}* cancels a running AI Platform job. If the job is already
finished, the command will not perform an operation and exit successfully.
"""
}


Cancel.detailed_help = _DETAILED_HELP



# -*- coding: utf-8 -*- #
# Copyright 2025 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Flags and helpers for the compute snapshot groups commands."""


from googlecloudsdk.command_lib.compute import completers as compute_completers
from googlecloudsdk.command_lib.compute import flags as compute_flags


def MakeSnapshotGroupArg(plural=False):
  return compute_flags.ResourceArgument(
      resource_name='snapshot-group',
      name='snapshot_group_name',
      completer=compute_completers.RoutesCompleter,
      plural=plural,
      global_collection='compute.snapshotGroups',
  )

SOURCE_INSTANT_SNAPSHOT_GROUP_ARG = compute_flags.ResourceArgument(
    resource_name='source instant snapshot group',
    name='--source-instant-snapshot-group',
    completer=compute_completers.InstantSnapshotGroupsCompleter,
    short_help="""
    The name or URL of the source instant snapshot group. If the name is provided, the instant snapshot group's zone
    or region must be specified with --source-instant-snapshot-group-zone or
    --source-instant-snapshot-group-region accordingly.
    To create a snapshot group from an instant snapshot group in a different project, specify the full path to the instant snapshot group.
    If the URL is provided, format should be:
    https://www.googleapis.com/compute/v1/projects/MY-PROJECT/zones/MY-ZONE/instantSnapshotGroups/MY-INSTANT-SNAPSHOT-GROUP
    """,
    zonal_collection='compute.instantSnapshotGroups',
    regional_collection='compute.regionInstantSnapshotGroups',
    required=True,
)




# -*- coding: utf-8 -*- #
# Copyright 2026 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base class for GCP resource handlers."""

import abc
import dataclasses
from typing import Any, Union

from apitools.base.protorpclite import messages
from apitools.base.py import encoding
from apitools.base.py import exceptions as apitools_exceptions
from googlecloudsdk.api_lib.util import apis
from googlecloudsdk.api_lib.util import waiter
from googlecloudsdk.command_lib.orchestration_pipelines import deployment_model
from googlecloudsdk.core import log
from googlecloudsdk.core import resources


IMPLICIT_LABEL_KEY = "orchestration_pipeline"
IMPLICIT_LABEL_VALUE = "true"


class GcpResourceHandler(abc.ABC):
  """An abstract handler for deploying a specific type of GCP resource."""

  api_name = None
  api_version = None
  api_prefix = "projects_locations"
  api_client_collection_path = None
  collection_name = None
  allowed_metadata = []

  # Documentation and Discoverability
  _parent_resource_type = None
  _documentation_uri = None

  def __init__(
      self,
      resource: deployment_model.ResourceModel,
      environment: deployment_model.EnvironmentModel,
      debug: bool = False,
  ):
    self.resource = resource
    self.environment = environment
    self.debug = debug
    self.client = self._get_client()
    self.messages = self._get_messages()
    self._validate_metadata()

  def get_labels_field_name(self) -> str | None:
    """Returns the name of the labels field, or None to disable.

    Subclasses can override this to specify a custom field name or disable
    implicit labeling by returning None. Returning "" (default) enables
    autodetection.
    """
    return ""

  def _detect_labels_field(self) -> str | None:
    """Detects the name of the labels field in the resource message."""
    override = self.get_labels_field_name()
    if override is None:
      return None
    if override:
      return override

    for field in self.resource_message_type.all_fields():
      if field.name == "labels":
        return "labels"
    return None

  def _build_labels_value(
      self, labels_value_type: Any, labels_dict: dict[str, str]
  ) -> Any:
    """Builds a LabelsValue message from a dictionary."""
    additional_properties = []
    for k, v in labels_dict.items():
      additional_properties.append(
          labels_value_type.AdditionalProperty(key=k, value=v)
      )
    return labels_value_type(additionalProperties=additional_properties)

  def _validate_metadata(self):
    """Validates that the resource metadata is applicable for this handler."""
    if not self.resource.metadata:
      return

    # attributes in MetadataModel that are not None
    used_metadata_keys = [
        f.name
        for f in dataclasses.fields(self.resource.metadata)
        if getattr(self.resource.metadata, f.name) is not None
    ]

    supported_metadata = {"location"}
    if self.allowed_metadata:
      supported_metadata.update(self.allowed_metadata)

    for key in used_metadata_keys:
      if key not in supported_metadata:
        msg = (
            f"Metadata field '{key}' is not applicable for resource type"
            f" '{self.resource.type}'."
        )
        if self.allowed_metadata:
          msg += f" Allowed metadata fields: {list(supported_metadata)}"
        else:
          msg += " This resource does not support any metadata fields."
        raise ValueError(msg)

    # Validate that nested resources have an explicit parent defined
    parts = self.resource.type.split(".")
    if len(parts) >= 3 and not self.resource.parent:
      parent_type = parts[-2]
      raise ValueError(
          f"Resource type '{self.resource.type}' represents a nested resource "
          f"and requires an explicit 'parent' (expected type: '{parent_type}') "
          "to be set in the resource configuration."
      )

  @property
  def _api_name(self) -> str:
    if self.api_name is not None:
      return self.api_name
    return self.resource.type.split(".")[0]

  @property
  def _api_version(self) -> str:
    if self.api_version is not None:
      return self.api_version
    # fallback to ResourceModel's api_version if available
    if self.resource.api_version is not None:
      return self.resource.api_version
    # otherwise find gcloud default
    return apis.ResolveVersion(self._api_name)

  @property
  def location(self) -> str:
    """Returns the effective location (region/zone/global) for the resource.

    Checks metadata.location first, falls back to environment region.
    """
    if self.resource.metadata.location is not None:
      return self.resource.metadata.location
    return self.environment.region

  def _get_location_path(self) -> str:
    """Returns the base path for the API location."""
    return f"projects/{self.environment.project}/locations/{self.location}"

  def _get_parent_path(self) -> str:
    """Returns the path to the parent resource."""
    location_path = self._get_location_path()
    if self.resource.parent:
      parts = self.resource.type.split(".")
      if len(parts) >= 3:
        parent_type = self._pluralize(parts[-2])
        parent_id = self.get_validated_parent_id()
        return f"{location_path}/{parent_type}/{parent_id}"
      else:
        raise ValueError(
            "Cannot infer parent type from resource type "
            f"'{self.resource.type}'"
        )
    return location_path

  def _get_client(self):
    return apis.GetClientInstance(self._api_name, self._api_version)

  def _get_messages(self):
    return apis.GetMessagesModule(self._api_name, self._api_version)

  def get_validated_parent_id(self) -> str:
    """Extracts and validates the ID from a parent string.

    Expected format is 'parentId'.

    Returns:
      The validated parent ID.

    Raises:
      ValueError: If the parent is missing or the format is invalid (e.g.
        contains '/').
    """
    if not self.resource.parent:
      raise ValueError(f"Parent must be specified for {self.resource.type}.")

    parent = self.resource.parent.strip()
    if "/" in parent:
      raise ValueError(
          f"Resource parent '{self.resource.parent}' must not contain '/'."
      )

    return parent

  def get_resource_id(self) -> str:
    """Returns the unique identifier for the resource."""
    return self.resource.name

  @classmethod
  def _pluralize(cls, word: str) -> str:
    """Returns the pluralized form of a word."""
    if word.endswith("y") and word[-2:-1].lower() not in "aeiouy":
      return word[:-1] + "ies"
    return word + "s"

  @property
  def _collection_name(self) -> str:
    if self.collection_name is not None:
      return self.collection_name
    return self._pluralize(self.resource.type.split(".")[-1])

  @property
  def _api_client_collection_path(self) -> str:
    """Returns the path to the API collection on the client."""
    if self.api_client_collection_path is not None:
      return self.api_client_collection_path

    parts = self.resource.type.split(".")[1:]
    pluralized_parts = []
    for i, part in enumerate(parts):
      if i == len(parts) - 1 and self.collection_name is not None:
        pluralized_parts.append(self.collection_name)
      else:
        pluralized_parts.append(self._pluralize(part))

    if self.api_prefix:
      return "_".join([self.api_prefix] + pluralized_parts)
    return "_".join(pluralized_parts)

  def _get_resource_name(self) -> str:
    """Returns the full resource name including location and collection."""
    return f"{self._get_parent_path()}/{self._collection_name}/{self.get_resource_id()}"

  def find_existing_resource(self) -> Any:
    """Finds the existing resource on GCP."""
    request = self.build_get_request()
    try:
      return self.get_get_method()(request)
    except apitools_exceptions.HttpNotFoundError:
      return None

  def get_parent_resource_type(self) -> str:
    """Returns the type of the parent resource, or None if top-level."""
    if self._parent_resource_type is not None:
      return self._parent_resource_type

    type_name = self.resource.type
    parts = type_name.split(".")
    if len(parts) > 2:
      parent_type = ".".join(parts[:-1])
    else:
      parent_type = None

    return parent_type

  def get_documentation_uri(self) -> str:
    """Returns a link to the REST API documentation for this resource."""
    if self._documentation_uri is not None:
      return self._documentation_uri

    try:
      resource_type = self.resource.type
      api_name = self.api_name if self.api_name else resource_type.split(".")[0]
      api_version = self.api_version
      if not api_version:
        api_version = apis.ResolveVersion(api_name)

      if self.api_client_collection_path:
        collection_path = self.api_client_collection_path
      else:
        parts = resource_type.split(".")[1:]
        pluralized_parts = []
        for i, part in enumerate(parts):
          if i == len(parts) - 1 and self.collection_name is not None:
            pluralized_parts.append(self.collection_name)
          else:
            pluralized_parts.append(self._pluralize(part))

        if self.api_prefix:
          collection_path = "_".join([self.api_prefix] + pluralized_parts)
        else:
          collection_path = "_".join(pluralized_parts)

      collection_path = collection_path.replace("_", ".")
      return (
          f"https://cloud.google.com/{api_name}/docs/reference/rest/"
          f"{api_version}/{collection_path}"
      )
    except Exception:  # pylint: disable=broad-exception-caught
      return "N/A"

  @abc.abstractmethod
  def build_get_request(self) -> messages.Message:
    """Builds the API request to get the resource."""

  def get_local_definition(self) -> dict[str, Any]:
    definition = self.resource.definition
    return definition.copy() if definition else {}

  @property
  def resource_message_type(self) -> type[messages.Message]:
    """Returns the apitools Message type for the resource."""
    # Try to infer from resource type
    # e.g. iam.serviceAccount -> ServiceAccount
    # e.g. dataform.repository.releaseConfig -> ReleaseConfig
    try:
      short_type = self.resource.type.split(".")[-1]
      # Uppercase first letter
      message_name = short_type[0].upper() + short_type[1:]
      if hasattr(self.messages, message_name):
        return getattr(self.messages, message_name)
    except (IndexError, AttributeError):
      pass

    raise NotImplementedError(
        f"Could not infer message type for {self.resource.type}, "
        "please override resource_message_type"
    )

  def to_resource_message(self, definition: dict[str, Any]) -> messages.Message:
    """Converts a dictionary definition to a resource message."""
    definition_copy = definition.copy()
    labels_field_name = self._detect_labels_field()
    if labels_field_name:
      labels = definition_copy.get(labels_field_name, {})
      if isinstance(labels, dict):
        labels_copy = labels.copy()
        labels_copy[IMPLICIT_LABEL_KEY] = IMPLICIT_LABEL_VALUE
        definition_copy[labels_field_name] = labels_copy
    return encoding.DictToMessage(definition_copy, self.resource_message_type)

  @abc.abstractmethod
  def build_create_request(
      self, resource_message: messages.Message
  ) -> messages.Message:
    """Builds the API request to create the resource."""

  @abc.abstractmethod
  def build_update_request(
      self,
      existing_resource: Union[messages.Message, dict[str, Any]],
      resource_message: messages.Message,
      changed_fields: list[str],
  ) -> messages.Message:
    """Builds the API request to update the resource."""

  @property
  def _api_client_collection(self) -> Any:
    return getattr(self.client, self._api_client_collection_path)

  def get_create_method(self) -> Any:
    """Returns the client method used to create the resource."""
    return self._api_client_collection.Create

  def get_update_method(self) -> Any:
    """Returns the client method used to update the resource."""
    return self._api_client_collection.Patch

  def get_get_method(self) -> Any:
    """Returns the client method used to get the resource."""
    return self._api_client_collection.Get

  def get_delete_method(self) -> Any:
    """Returns the client method used to delete the resource."""
    return self._api_client_collection.Delete

  def compare(
      self, existing_resource: Any, local_definition: dict[str, Any]
  ) -> list[str]:
    """Compares existing resource with local definition."""
    existing_dict = existing_resource
    # If the resource is an apitools Message, convert it to a dict.
    if not isinstance(existing_resource, dict):
      try:
        existing_dict = encoding.MessageToDict(existing_resource)
      except Exception:  # pylint: disable=broad-except
        # Fallback for non-Message objects (e.g. in tests/mocks)
        if hasattr(existing_resource, "__dict__"):
          existing_dict = existing_resource.__dict__

    diffs = self._compare_recursive(existing_dict, local_definition, "")

    if diffs:
      labels_field_name = self._detect_labels_field()
      if labels_field_name and labels_field_name not in diffs:
        existing_labels = existing_dict.get(labels_field_name, {})
        if isinstance(existing_labels, dict):
          if existing_labels.get(IMPLICIT_LABEL_KEY) != IMPLICIT_LABEL_VALUE:
            diffs.append(labels_field_name)

    return diffs

  def _compare_recursive(
      self, existing: Any, local: Any, prefix: str
  ) -> list[str]:
    """Recursively compares two objects and returns changed fields."""
    diffs = []
    is_local_dict = isinstance(local, dict)
    is_existing_dict = isinstance(existing, dict)

    if is_local_dict:
      if not is_existing_dict:
        # If existing is not a dict, the whole tree at 'prefix' matches nothing
        # useful in structure, so we report the prefix itself (or local keys).
        return [prefix] if prefix else list(local.keys())

      for k, v in local.items():
        next_prefix = f"{prefix}.{k}" if prefix else k
        if k not in existing:
          diffs.append(next_prefix)
        else:
          diffs.extend(self._compare_recursive(existing[k], v, next_prefix))
    elif isinstance(local, list) and isinstance(existing, list):
      if len(local) != len(existing):
        return [prefix] if prefix else []
      for ext_item, loc_item in zip(existing, local):
        item_diffs = self._compare_recursive(ext_item, loc_item, prefix)
        if item_diffs:
          return [prefix] if prefix else []
    else:
      # Primitive
      if isinstance(existing, str) and isinstance(local, str):
        if existing != local:
          # Permit API URLs to match partial configuration paths natively
          if not (existing.startswith("https://") and existing.endswith(local)):
            return [prefix] if prefix else []
      elif str(existing) != str(local) and existing != local:
        return [prefix] if prefix else []

    return diffs

  def wait_for_operation(self, api_response: Any) -> Any:
    """Waits for long running operation if applicable.

    The default implementation handles resources that don't return LROs.
    It returns the api_response as is.
    Handlers for resources that return LROs should override this method.

    Args:
      api_response: The operation to wait for, or the resource if no LRO.

    Returns:
      The completed operation or the originally returned resource.
    """

    if type(api_response).__name__ == "Operation":
      if getattr(api_response, "done", False):
        return api_response

      op_name = getattr(api_response, "name", "unknown")
      log.status.Print(f"     Waiting for operation {op_name} to complete...")

      # Determine collection for the operation based on the current api client
      operation_collection = f"{self.api_prefix}_operations".replace("_", ".")
      if not operation_collection.startswith(self._api_name):
        operation_collection = f"{self._api_name}.{operation_collection}"

      try:
        operation_ref = resources.REGISTRY.Parse(
            api_response.name,
            collection=operation_collection,
        )
      except resources.Error:
        # Fallback if parse fails
        return api_response

      operation_service_name = f"{self.api_prefix}_operations"
      if not hasattr(self.client, operation_service_name):
        raise ValueError(
            f"Operation service {operation_service_name} not found on client "
            f"for {self._api_name}."
        )
      ops_service = getattr(self.client, operation_service_name)

      poller = waiter.CloudOperationPollerNoResources(ops_service)
      try:
        api_response = waiter.WaitFor(
            poller,
            operation_ref,
            f"Waiting for {self._api_name} operation",
        )
      except waiter.TimeoutError:
        log.error(f"Timed out waiting for {self._api_name} operation.")
    return api_response

  def get_success_deployment_message(self, api_response: Any) -> str:  # pylint: disable=unused-argument
    """Returns the identifier to print after successful resource deployment.

    Args:
      api_response: The API response returned by wait_for_operation or create.

    By default this returns the logical name defined in the local configuration.
    Subclasses should override this if the generated API identifier is required
    by users to track the resource (e.g. BQ DTS config IDs).
    """
    return self.resource.name

  def post_deploy(self, api_response: Any, created: bool) -> None:
    """Executes actions after a successful deployment.

    Args:
      api_response: The API response returned by the deployment operation.
      created: True if the resource was created, False if it was updated.
    """
    pass



# -*- coding: utf-8 -*- #
# Copyright 2026 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Describe command for Threat Prevention profiles."""


from googlecloudsdk.api_lib.network_security.security_profiles import tpp_api
from googlecloudsdk.calliope import base
# sp_flags handles both Organization and Project-level resource arguments.
from googlecloudsdk.command_lib.network_security import sp_flags

_DETAILED_HELP = {
    'DESCRIPTION': """
          Show details of a Threat Prevention Security Profile.

          For more examples, refer to the EXAMPLES section below.
        """,
    'EXAMPLES': """
            To show details of a Threat Prevention security profile named `my-security-profile` in an organization, run:

            $ {command} my-security-profile --organization=1234 --location=global

            To show details of a Threat Prevention security profile named `my-security-profile` in a project, run:

            $ {command} my-security-profile --project=my-project --location=global
        """,
}

_PROJECT_SCOPE_SUPPORTED_TRACKS = (
    base.ReleaseTrack.ALPHA,
)


@base.DefaultUniverseOnly
@base.ReleaseTracks(
    base.ReleaseTrack.ALPHA, base.ReleaseTrack.BETA, base.ReleaseTrack.GA
)
class Describe(base.DescribeCommand):
  """Describe a Threat Prevention Security Profile."""

  @classmethod
  def Args(cls, parser):
    project_scope_supported = (
        cls.ReleaseTrack() in _PROJECT_SCOPE_SUPPORTED_TRACKS
    )
    sp_flags.AddSecurityProfileResource(
        parser, cls.ReleaseTrack(), project_scope_supported
    )

  def Run(self, args):
    result = args.CONCEPTS.security_profile.Parse()
    security_profile = result.result

    project_scoped = (
        result.concept_type.name
        == sp_flags.PROJECT_SECURITY_PROFILE_RESOURCE_COLLECTION
    )

    client = tpp_api.Client(self.ReleaseTrack(), project_scoped)

    # Returning the full SecurityProfile resource.
    return client.GetSecurityProfile(security_profile.RelativeName())


Describe.detailed_help = _DETAILED_HELP



# -*- coding: utf-8 -*- #
# Copyright 2018 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Provides helper methods for dealing with JSON files for Bigtable IAM."""


from googlecloudsdk.api_lib.bigtable import instances
from googlecloudsdk.api_lib.bigtable import util
from googlecloudsdk.command_lib.iam import iam_util


def SetInstanceIamPolicy(instance_ref, policy):
  """Sets the IAM policy on an instance."""
  msgs = util.GetAdminMessages()
  policy = iam_util.ParsePolicyFile(policy, msgs.Policy)
  return instances.SetIamPolicy(instance_ref, policy)



# -*- coding: utf-8 -*- #
# Copyright 2014 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command for snapshotting disks."""


from googlecloudsdk.api_lib.compute import base_classes
from googlecloudsdk.api_lib.compute import csek_utils
from googlecloudsdk.api_lib.compute import name_generator
from googlecloudsdk.api_lib.compute.operations import poller
from googlecloudsdk.api_lib.util import waiter
from googlecloudsdk.calliope import arg_parsers
from googlecloudsdk.calliope import base
from googlecloudsdk.calliope import exceptions
from googlecloudsdk.command_lib.compute import flags
from googlecloudsdk.command_lib.compute.disks import flags as disks_flags
from googlecloudsdk.command_lib.compute.snapshots import flags as snap_flags
from googlecloudsdk.command_lib.util.args import labels_util
from googlecloudsdk.core import exceptions as core_exceptions
from googlecloudsdk.core import log
from googlecloudsdk.core import properties
from six.moves import zip


DETAILED_HELP = {
    'brief': 'Create snapshots of Compute Engine persistent disks.',
    'DESCRIPTION': """
        *{command}* creates snapshots of persistent disks. Snapshots are useful
        for backing up data, copying a persistent disk, and even, creating a
        custom image. Snapshots can be created from persistent disks even while
        they are attached to running instances. Once created, snapshots may be
        managed (listed, deleted, etc.) via `gcloud compute snapshots`.

        Refer to the Snapshot best practices guide. https://cloud.google.com/compute/docs/disks/snapshot-best-practices

        {command} waits until the operation returns a status of `READY` or
        `FAILED`, or reaches the maximum timeout, and returns the last known
        details of the snapshot.

        Note: To create snapshots, the following IAM permissions are necessary
        ``compute.disks.createSnapshot'', ``compute.snapshots.create'',
        ``compute.snapshots.get'', and ``compute.zoneOperations.get''.
        """,
    'EXAMPLES': """
        To create a snapshot named `snapshot-test` of a persistent disk named `test`
        in zone `us-central1-a`, run:

          $ {command} test --zone=us-central1-a --snapshot-names=snapshot-test --description="This is an example snapshot"
    """
}


def _CommonArgs(parser):
  """Add parser arguments common to all tracks."""
  SnapshotDisks.disks_arg.AddArgument(parser)

  parser.add_argument(
      '--description',
      help=('Text to describe the snapshots being created.'))
  parser.add_argument(
      '--snapshot-names',
      type=arg_parsers.ArgList(min_length=1),
      metavar='SNAPSHOT_NAME',
      help="""\
      Names to assign to the created snapshots. Without this option, the
      name of each snapshot will be a random 12-character alphanumeric
      string that starts with a letter. The values of
      this option run parallel to the disks specified. For example,

          {command} my-disk-1 my-disk-2 my-disk-3 --snapshot-names snapshot-1,snapshot-2,snapshot-3

      will result in `my-disk-1` being snapshotted as
      `snapshot-1`, `my-disk-2` as `snapshot-2`, and so on. The name must match
      the `(?:[a-z](?:[-a-z0-9]{0,61}[a-z0-9])?)` regular expression, which
      means it must start with an alphabetic character followed by one or more
      alphanumeric characters or dashes. The name must not exceed 63 characters
      and must not contain special symbols. All characters must be lowercase.
      """)
  snap_flags.AddChainArg(parser)
  flags.AddGuestFlushFlag(parser, 'snapshot')
  flags.AddStorageLocationFlag(parser, 'snapshot')
  csek_utils.AddCsekKeyArgs(parser, flags_about_creation=False)

  base.ASYNC_FLAG.AddToParser(parser)


@base.ReleaseTracks(base.ReleaseTrack.GA, base.ReleaseTrack.PREVIEW)
class SnapshotDisks(base.SilentCommand):
  """Create snapshots of Google Compute Engine persistent disks."""

  @classmethod
  def Args(cls, parser):
    SnapshotDisks.disks_arg = disks_flags.MakeDiskArg(plural=True)
    labels_util.AddCreateLabelsFlags(parser)
    _CommonArgs(parser)

  def Run(self, args):
    return self._Run(args)

  def _Run(self, args):
    """Returns a list of requests necessary for snapshotting disks."""
    holder = base_classes.ComputeApiHolder(self.ReleaseTrack())

    disk_refs = SnapshotDisks.disks_arg.ResolveAsResource(
        args, holder.resources,
        scope_lister=flags.GetDefaultScopeLister(holder.client))
    if args.snapshot_names:
      if len(disk_refs) != len(args.snapshot_names):
        raise exceptions.InvalidArgumentException(
            '--snapshot-names',
            '[--snapshot-names] must have the same number of values as disks '
            'being snapshotted.')
      snapshot_names = args.snapshot_names
    else:
      # Generates names like "d52jsqy3db4q".
      snapshot_names = [name_generator.GenerateRandomName()
                        for _ in disk_refs]

    snapshot_refs = [
        holder.resources.Parse(
            snapshot_name,
            params={
                'project': properties.VALUES.core.project.GetOrFail,
            },
            collection='compute.snapshots')
        for snapshot_name in snapshot_names]

    client = holder.client.apitools_client
    messages = holder.client.messages

    requests = []

    for disk_ref, snapshot_ref in zip(disk_refs, snapshot_refs):
      csek_keys = csek_utils.CsekKeyStore.FromArgs(args, True)
      snapshot_key_or_none = csek_utils.MaybeLookupKeyMessage(
          csek_keys, snapshot_ref, client)
      disk_key_or_none = csek_utils.MaybeLookupKeyMessage(
          csek_keys, disk_ref, client)

      snapshot_message = messages.Snapshot(
          name=snapshot_ref.Name(),
          description=args.description,
          snapshotEncryptionKey=snapshot_key_or_none,
          sourceDiskEncryptionKey=disk_key_or_none,
          chainName=args.chain_name)

      if (hasattr(args, 'storage_location') and
          args.IsSpecified('storage_location')):
        snapshot_message.storageLocations = [args.storage_location]
      if (hasattr(args, 'labels') and args.IsSpecified('labels')):
        snapshot_message.labels = labels_util.ParseCreateArgs(
            args, messages.Snapshot.LabelsValue)

      if disk_ref.Collection() == 'compute.disks':
        request = messages.ComputeDisksCreateSnapshotRequest(
            disk=disk_ref.Name(),
            snapshot=snapshot_message,
            project=disk_ref.project,
            zone=disk_ref.zone,
            guestFlush=args.guest_flush)
        requests.append((client.disks, 'CreateSnapshot', request))
      elif disk_ref.Collection() == 'compute.regionDisks':
        request = messages.ComputeRegionDisksCreateSnapshotRequest(
            disk=disk_ref.Name(),
            snapshot=snapshot_message,
            project=disk_ref.project,
            region=disk_ref.region)
        if hasattr(request, 'guestFlush'):  # only available in alpha API
          guest_flush = getattr(args, 'guest_flush', None)
          if guest_flush is not None:
            request.guestFlush = guest_flush
        requests.append((client.regionDisks, 'CreateSnapshot', request))

    errors_to_collect = []
    responses = holder.client.AsyncRequests(requests, errors_to_collect)
    for r in responses:
      err = getattr(r, 'error', None)
      if err:
        errors_to_collect.append(poller.OperationErrors(err.errors))
    if errors_to_collect:
      raise core_exceptions.MultiError(errors_to_collect)

    operation_refs = [holder.resources.Parse(r.selfLink) for r in responses]

    if args.async_:
      for operation_ref in operation_refs:
        log.status.Print('Disk snapshot in progress for [{}].'
                         .format(operation_ref.SelfLink()))
      log.status.Print('Use [gcloud compute operations describe URI] command '
                       'to check the status of the operation(s).')
      return responses

    operation_poller = poller.BatchPoller(
        holder.client, client.snapshots, snapshot_refs)
    return waiter.WaitFor(
        operation_poller, poller.OperationBatch(operation_refs),
        'Creating snapshot(s) {0}'
        .format(', '.join(s.Name() for s in snapshot_refs)),
        max_wait_ms=None
    )


@base.ReleaseTracks(base.ReleaseTrack.BETA)
class SnapshotDisksBeta(SnapshotDisks):
  """Create snapshots of Google Compute Engine persistent disks beta."""

  @classmethod
  def Args(cls, parser):
    SnapshotDisks.disks_arg = disks_flags.MakeDiskArg(plural=True)
    labels_util.AddCreateLabelsFlags(parser)
    _CommonArgs(parser)


@base.ReleaseTracks(base.ReleaseTrack.ALPHA)
class SnapshotDisksAlpha(SnapshotDisksBeta):
  """Create snapshots of Google Compute Engine persistent disks alpha."""

  @classmethod
  def Args(cls, parser):
    SnapshotDisks.disks_arg = disks_flags.MakeDiskArg(plural=True)
    labels_util.AddCreateLabelsFlags(parser)
    _CommonArgs(parser)

  def Run(self, args):
    return self._Run(args)


SnapshotDisks.detailed_help = DETAILED_HELP



# -*- coding: utf-8 -*- #
# Copyright 2022 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Client for interacting with Storage Insights."""


import collections
from typing import Sequence, Optional

from apitools.base.py import list_pager

from googlecloudsdk.api_lib.storage import errors
from googlecloudsdk.api_lib.storage.gcs_json import client as gcs_json_client
from googlecloudsdk.api_lib.util import apis as core_apis
from googlecloudsdk.core import properties


# Backend has a limit of 500.
PAGE_SIZE = 500
_CSV_PARQUET_ERROR_MESSGE = 'CSV options cannot be set with parquet.'


def _get_unescaped_ascii(string):
  """Returns the ASCII string unescaping any escaped characters."""
  return string.encode('ascii').decode(
      'unicode-escape') if string is not None else None


def _get_parent_string(
    project,
    location,
):
  return 'projects/{}/locations/{}'.format(project, location.lower())


def _get_parent_string_from_bucket(bucket):
  gcs_client = gcs_json_client.JsonClient()
  bucket_resource = gcs_client.get_bucket(bucket)
  return _get_parent_string(
      bucket_resource.metadata.projectNumber,
      bucket_resource.metadata.location.lower(),
  )


# Tuple to hold the CsvOptions or ParquetOptions. Only one of the fields should
# be set.
ReportFormatOptions = collections.namedtuple(
    'ReportFormatOptions', ('csv', 'parquet')
)


class InsightsApi:
  """Client for Storage Insights API."""

  def __init__(self):
    super(InsightsApi, self).__init__()
    self.client = core_apis.GetClientInstance('storageinsights', 'v1')
    self.messages = core_apis.GetMessagesModule('storageinsights', 'v1')

  def create_dataset_config(
      self,
      dataset_config_name: str,
      location: str,
      destination_project: str,
      organization_number: int,
      retention_period: int,
      activity_data_retention_period: Optional[int] = None,
      organization_scope: bool = False,
      source_projects_list: Optional[Sequence[int]] = None,
      source_folders_list: Optional[Sequence[int]] = None,
      include_buckets_prefix_regex_list: Optional[Sequence[str]] = None,
      exclude_buckets_prefix_regex_list: Optional[Sequence[str]] = None,
      include_buckets_name_list: Optional[Sequence[str]] = None,
      exclude_buckets_name_list: Optional[Sequence[str]] = None,
      include_source_locations: Optional[Sequence[str]] = None,
      exclude_source_locations: Optional[Sequence[str]] = None,
      auto_add_new_buckets: bool = False,
      identity_type: str = None,
      description: str = None,
  ):
    """Creates a dataset config.

    Args:
      dataset_config_name: Name for the dataset config being created.
      location: The location where insights data will be stored in a GCS
        managed BigQuery instance.
      destination_project: The project in which the dataset config is
        being created and by extension the insights data will be stored.
      organization_number: Organization number of the organization to
        which all source projects / folders must belong.
      retention_period: No of days for which insights data is to be
        retained in BigQuery instance.
      activity_data_retention_period: No of days for which activity data is to
        be retained in BigQuery instance. If not set, retention period will be
        used by the API.
      organization_scope: If True, Insights data will be collected for
        all resources in the given organization.
      source_projects_list: List of source project numbers. Insights
        data is to be collected for buckets that belong to these projects.
      source_folders_list: List of source folder numbers. Insights
        data is to be collected for buckets that belong to these folders.
      include_buckets_prefix_regex_list: List of bucket prefix regex
        patterns which are to be included for insights processing from the
        source projects. We can either use included or excluded bucket
        parameters.
      exclude_buckets_prefix_regex_list: List of bucket prefix regex
        patterns which are to be excluded from insights processing from the
        source projects. We can either use included or excluded bucket
        parameters.
      include_buckets_name_list: List of bucket names which are to
        be included for insights processing from the source projects. We can
        either use included or excluded bucket parameters.
      exclude_buckets_name_list: List of bucket names which are to
        be excluded from insights processing from the source projects. We can
        either use included or excluded bucket parameters.
      include_source_locations: List of bucket locations which are
        to be included for insights processing from the source projects. We can
        either use included or excluded location parameters.
      exclude_source_locations: List of bucket locations which are
        to be excluded from insights processing from the source projects. We can
        either use included or excluded location parameters.
      auto_add_new_buckets: If True, auto includes any new buckets added
        to source projects that satisfy the include/exclude criterias.
      identity_type: Option for how permissions need to be setup for a
        given dataset config. Default option is IDENTITY_TYPE_PER_CONFIG.
      description: Human readable description text for the given dataset
        config.

    Returns:
      An instance of Operation message
    """
    if identity_type is not None:
      identity_type_enum = self.messages.Identity.TypeValueValuesEnum(
          identity_type.upper()
      )
      identity_type = self.messages.Identity(type=identity_type_enum)
    else:
      identity_type = self.messages.Identity(
          type=self.messages.Identity.TypeValueValuesEnum.IDENTITY_TYPE_PER_CONFIG
      )

    source_projects = self.messages.SourceProjects(
        projectNumbers=source_projects_list
    ) if source_projects_list else None

    source_folders = self.messages.SourceFolders(
        folderNumbers=source_folders_list
    ) if source_folders_list else None

    dataset_config = self.messages.DatasetConfig(
        description=description,
        identity=identity_type,
        includeNewlyCreatedBuckets=auto_add_new_buckets,
        name=dataset_config_name,
        organizationNumber=organization_number,
        retentionPeriodDays=retention_period,
        sourceProjects=source_projects,
        sourceFolders=source_folders,
        organizationScope=organization_scope if organization_scope else None,
    )

    if activity_data_retention_period is not None:
      dataset_config.activityDataRetentionPeriodDays = (
          activity_data_retention_period
      )

    # Exclude and Include options are marked as mutex flags,
    # hence we can make the below assumption about only being available
    if exclude_buckets_name_list or exclude_buckets_prefix_regex_list:
      excluded_storage_buckets = [
          self.messages.CloudStorageBucket(bucketName=excluded_name)
          for excluded_name in exclude_buckets_name_list or []
      ]
      excluded_storage_buckets += [
          self.messages.CloudStorageBucket(bucketPrefixRegex=excluded_regex)
          for excluded_regex in exclude_buckets_prefix_regex_list or []
      ]
      dataset_config.excludeCloudStorageBuckets = (
          self.messages.CloudStorageBuckets(
              cloudStorageBuckets=excluded_storage_buckets
          )
      )

    if include_buckets_name_list or include_buckets_prefix_regex_list:
      included_storage_buckets = [
          self.messages.CloudStorageBucket(bucketName=included_name)
          for included_name in include_buckets_name_list or []
      ]
      included_storage_buckets += [
          self.messages.CloudStorageBucket(bucketPrefixRegex=included_regex)
          for included_regex in include_buckets_prefix_regex_list or []
      ]
      dataset_config.includeCloudStorageBuckets = (
          self.messages.CloudStorageBuckets(
              cloudStorageBuckets=included_storage_buckets
          )
      )

    if exclude_source_locations:
      dataset_config.excludeCloudStorageLocations = (
          self.messages.CloudStorageLocations(
              locations=exclude_source_locations
          )
      )

    if include_source_locations:
      dataset_config.includeCloudStorageLocations = (
          self.messages.CloudStorageLocations(
              locations=include_source_locations
          )
      )

    create_request = self.messages.StorageinsightsProjectsLocationsDatasetConfigsCreateRequest(
        datasetConfig=dataset_config,
        datasetConfigId=dataset_config_name,
        parent=_get_parent_string(destination_project, location),
    )
    return self.client.projects_locations_datasetConfigs.Create(create_request)

  def create_dataset_config_link(self, dataset_config_relative_name):
    """Creates the dataset config link."""
    request = self.messages.StorageinsightsProjectsLocationsDatasetConfigsLinkDatasetRequest(
        name=dataset_config_relative_name
    )
    return self.client.projects_locations_datasetConfigs.LinkDataset(request)

  def delete_dataset_config(self, dataset_config_relative_name):
    """Deletes the dataset config."""
    request = self.messages.StorageinsightsProjectsLocationsDatasetConfigsDeleteRequest(
        name=dataset_config_relative_name
    )
    return self.client.projects_locations_datasetConfigs.Delete(request)

  def delete_dataset_config_link(self, dataset_config_relative_name):
    """Deletes the dataset config link."""
    request = self.messages.StorageinsightsProjectsLocationsDatasetConfigsUnlinkDatasetRequest(
        name=dataset_config_relative_name
    )
    return self.client.projects_locations_datasetConfigs.UnlinkDataset(request)

  def get_dataset_config(self, dataset_config_relative_name):
    """Gets the dataset config."""
    return self.client.projects_locations_datasetConfigs.Get(
        self.messages.StorageinsightsProjectsLocationsDatasetConfigsGetRequest(
            name=dataset_config_relative_name
        )
    )

  def list_dataset_config(self, location=None, page_size=None):
    """Lists the dataset configs.

    Args:
      location (str): The location where insights data will be stored in a GCS
        managed BigQuery instance.
      page_size (int|None): Number of items per request to be returned.

    Returns:
      List of dataset configs.
    """
    if location is not None:
      parent = _get_parent_string(
          properties.VALUES.core.project.Get(), location
      )
    else:
      parent = _get_parent_string(properties.VALUES.core.project.Get(), '-')

    return list_pager.YieldFromList(
        self.client.projects_locations_datasetConfigs,
        self.messages.StorageinsightsProjectsLocationsDatasetConfigsListRequest(
            parent=parent
        ),
        batch_size=page_size if page_size is not None else PAGE_SIZE,
        batch_size_attribute='pageSize',
        field='datasetConfigs',
    )

  def _get_dataset_config_update_mask(
      self,
      organization_scope: bool = None,
      source_projects: Optional[Sequence[int]] = None,
      source_folders: Optional[Sequence[int]] = None,
      include_buckets_name_list: Optional[Sequence[str]] = None,
      include_buckets_prefix_regex_list: Optional[Sequence[str]] = None,
      exclude_buckets_name_list: Optional[Sequence[str]] = None,
      exclude_buckets_prefix_regex_list: Optional[Sequence[str]] = None,
      include_source_locations: Optional[Sequence[str]] = None,
      exclude_source_locations: Optional[Sequence[str]] = None,
      auto_add_new_buckets: bool = None,
      retention_period: int = None,
      activity_data_retention_period: Optional[int] = None,
      description: str = None,
  ):
    """Returns the update_mask list."""
    update_mask = []

    mask_conditions = [
        ('organizationScope', organization_scope),
        ('sourceProjects', source_projects is not None),
        ('sourceFolders', source_folders is not None),
        (
            'includeCloudStorageBuckets',
            include_buckets_name_list is not None
            or include_buckets_prefix_regex_list is not None,
        ),
        (
            'excludeCloudStorageBuckets',
            exclude_buckets_name_list is not None
            or exclude_buckets_prefix_regex_list is not None,
        ),
        ('includeCloudStorageLocations', include_source_locations is not None),
        ('excludeCloudStorageLocations', exclude_source_locations is not None),
        ('includeNewlyCreatedBuckets', auto_add_new_buckets is not None),
        ('retentionPeriodDays', retention_period is not None),
        (
            'activityDataRetentionPeriodDays',
            activity_data_retention_period is not None,
        ),
        ('description', description is not None),
    ]

    for mask, condition in mask_conditions:
      if condition:
        update_mask.append(mask)

    return update_mask

  def update_dataset_config(
      self,
      dataset_config_relative_name,
      organization_scope: bool = None,
      source_projects_list: Optional[Sequence[int]] = None,
      source_folders_list: Optional[Sequence[int]] = None,
      include_buckets_name_list: Optional[Sequence[str]] = None,
      include_buckets_prefix_regex_list: Optional[Sequence[str]] = None,
      exclude_buckets_name_list: Optional[Sequence[str]] = None,
      exclude_buckets_prefix_regex_list: Optional[Sequence[str]] = None,
      include_source_locations: Optional[Sequence[str]] = None,
      exclude_source_locations: Optional[Sequence[str]] = None,
      auto_add_new_buckets: bool = None,
      retention_period: int = None,
      activity_data_retention_period: Optional[int] = None,
      description: str = None,
  ):
    """Updates the dataset config.

    Args:
      dataset_config_relative_name: The relative name of the dataset
        config to be modified.
      organization_scope: If True, Insights data will be collected for
        all resources in the given organization. If this field is Empty/None,
        Either the sourceProjects or the sourceFolders must be set to update
        the scope.
      source_projects_list: List of source project numbers. Insights
        data is to be collected for buckets that belong to these projects. If
        this field is Empty/None, Either the organizationScope will be True or
        the sourceFolders must be set to update the scope.
      source_folders_list: List of source folder IDs. Insights
        data is to be collected for buckets that belong to these organisational
        folders. If this field is Empty/None, Either the organizationScope will
        be True or the sourceProjects must be set to update the scope.
      include_buckets_name_list: Optional list of bucket names which
        are to be included for insights processing from the source projects. Can
        either use included or excluded bucket parameters.
      include_buckets_prefix_regex_list: Optional list of bucket
        prefix regex patterns which are to be included for insights processing
        from the source projects. Can either use included or excluded bucket
        parameters.
      exclude_buckets_name_list: Optional list of bucket names which
        are to be excluded from insights processing from the source projects.
        Can either use included or excluded bucket parameters.
      exclude_buckets_prefix_regex_list: Optional list of bucket
        prefix regex patterns which are to be excluded from insights processing
        from the source projects. Can either use included or excluded bucket
        parameters.
      include_source_locations: Optional list of bucket locations
        which are to be included for insights processing from the source
        projects. Can either use included or excluded location parameters.
      exclude_source_locations: Optional list of bucket locations
        which are to be excluded from insights processing from the source
        projects. Can either use included or excluded location parameters.
      auto_add_new_buckets: If True, auto includes any new buckets added
        to source projects that satisfy the include/exclude criterias.
      retention_period: No of days for which insights data is to be
        retained in BigQuery instance.
      activity_data_retention_period: No of days for which activity data is to
        be retained in BigQuery instance.
      description: Human readable description text for the given dataset
        config.

    Returns:
      An instance of Operation message.
    """

    # Only the fields present in the mask will be updated.
    update_mask = self._get_dataset_config_update_mask(
        organization_scope=organization_scope,
        source_projects=source_projects_list,
        source_folders=source_folders_list,
        include_buckets_name_list=include_buckets_name_list,
        include_buckets_prefix_regex_list=include_buckets_prefix_regex_list,
        exclude_buckets_name_list=exclude_buckets_name_list,
        exclude_buckets_prefix_regex_list=exclude_buckets_prefix_regex_list,
        include_source_locations=include_source_locations,
        exclude_source_locations=exclude_source_locations,
        auto_add_new_buckets=auto_add_new_buckets,
        retention_period=retention_period,
        activity_data_retention_period=activity_data_retention_period,
        description=description,
    )

    if not update_mask:
      raise errors.InsightApiError(
          'Nothing to update for dataset config: {}'.format(
              dataset_config_relative_name
          )
      )

    source_projects = self.messages.SourceProjects(
        projectNumbers=source_projects_list
    ) if source_projects_list is not None else None

    source_folders = self.messages.SourceFolders(
        folderNumbers=source_folders_list
    ) if source_folders_list is not None else None

    dataset_config = self.messages.DatasetConfig(
        retentionPeriodDays=retention_period,
        description=description,
        sourceProjects=source_projects,
        sourceFolders=source_folders,
        includeNewlyCreatedBuckets=auto_add_new_buckets,
    )

    if organization_scope:
      dataset_config.organizationScope = True

    if activity_data_retention_period is not None:
      dataset_config.activityDataRetentionPeriodDays = (
          activity_data_retention_period
      )

    # Exclude and Include options are marked as mutex flags,
    # hence can make the below assumption about only one of the options being
    # available.
    if exclude_buckets_name_list or exclude_buckets_prefix_regex_list:
      excluded_storage_buckets = [
          self.messages.CloudStorageBucket(bucketName=excluded_name)
          for excluded_name in exclude_buckets_name_list or []
      ]
      excluded_storage_buckets += [
          self.messages.CloudStorageBucket(bucketPrefixRegex=excluded_regex)
          for excluded_regex in exclude_buckets_prefix_regex_list or []
      ]
      dataset_config.excludeCloudStorageBuckets = (
          self.messages.CloudStorageBuckets(
              cloudStorageBuckets=excluded_storage_buckets
          )
      )
    elif include_buckets_name_list or include_buckets_prefix_regex_list:
      included_storage_buckets = [
          self.messages.CloudStorageBucket(bucketName=included_name)
          for included_name in include_buckets_name_list or []
      ]
      included_storage_buckets += [
          self.messages.CloudStorageBucket(bucketPrefixRegex=included_regex)
          for included_regex in include_buckets_prefix_regex_list or []
      ]
      dataset_config.includeCloudStorageBuckets = (
          self.messages.CloudStorageBuckets(
              cloudStorageBuckets=included_storage_buckets
          )
      )

    if exclude_source_locations:
      dataset_config.excludeCloudStorageLocations = (
          self.messages.CloudStorageLocations(
              locations=exclude_source_locations
          )
      )

    if include_source_locations:
      dataset_config.includeCloudStorageLocations = (
          self.messages.CloudStorageLocations(
              locations=include_source_locations
          )
      )

    request = self.messages.StorageinsightsProjectsLocationsDatasetConfigsPatchRequest(
        name=dataset_config_relative_name,
        datasetConfig=dataset_config,
        updateMask=','.join(update_mask),
    )
    return self.client.projects_locations_datasetConfigs.Patch(request)

  def _get_report_format_options(
      self, csv_separator, csv_delimiter, csv_header, parquet
  ):
    """Returns ReportFormatOptions instance."""
    if parquet:
      parquet_options = self.messages.ParquetOptions()
      if csv_header or csv_delimiter or csv_separator:
        raise errors.GcsApiError('CSV options cannot be set with parquet.')
      csv_options = None
    else:
      parquet_options = None
      unescaped_separator = _get_unescaped_ascii(csv_separator)
      csv_options = self.messages.CSVOptions(
          delimiter=csv_delimiter,
          headerRequired=csv_header,
          recordSeparator=unescaped_separator,
      )
    return ReportFormatOptions(
        csv=csv_options,
        parquet=parquet_options,
    )

  def create_inventory_report(
      self,
      source_bucket,
      destination_url,
      metadata_fields=None,
      start_date=None,
      end_date=None,
      frequency=None,
      csv_separator=None,
      csv_delimiter=None,
      csv_header=None,
      parquet=None,
      display_name=None,
  ):
    """Creates a report config.

    Args:
      source_bucket (str): Source bucket name for which reports will be
        generated.
      destination_url (storage_url.CloudUrl): The destination url where the
        generated reports will be stored.
      metadata_fields (list[str]): Fields to be included in the report.
      start_date (datetime.datetime.date): The date to start generating reports.
      end_date (datetime.datetime.date): The date after which to stop generating
        reports.
      frequency (str): Can be either DAILY or WEEKLY.
      csv_separator (str): The character used to separate the records in the
        CSV file.
      csv_delimiter (str): The delimiter that separates the fields in the CSV
        file.
      csv_header (bool): If True, include the headers in the CSV file.
      parquet (bool): If True, set the parquet options.
      display_name (str): Display name for the report config.

    Returns:
      The created ReportConfig object.
    """
    frequency_options = self.messages.FrequencyOptions(
        startDate=self.messages.Date(
            year=start_date.year, month=start_date.month, day=start_date.day),
        endDate=self.messages.Date(
            year=end_date.year, month=end_date.month, day=end_date.day),
        frequency=getattr(
            self.messages.FrequencyOptions.FrequencyValueValuesEnum,
            frequency.upper()))
    object_metadata_report_options = self.messages.ObjectMetadataReportOptions(
        metadataFields=metadata_fields,
        storageDestinationOptions=self.messages.CloudStorageDestinationOptions(
            bucket=destination_url.bucket_name,
            destinationPath=destination_url.resource_name),
        storageFilters=self.messages.CloudStorageFilters(
            bucket=source_bucket))

    report_format_options = self._get_report_format_options(
        csv_separator, csv_delimiter, csv_header, parquet)

    report_config = self.messages.ReportConfig(
        csvOptions=report_format_options.csv,
        parquetOptions=report_format_options.parquet,
        displayName=display_name,
        frequencyOptions=frequency_options,
        objectMetadataReportOptions=object_metadata_report_options)
    create_request = self.messages.StorageinsightsProjectsLocationsReportConfigsCreateRequest(
        parent=_get_parent_string_from_bucket(source_bucket),
        reportConfig=report_config)
    return self.client.projects_locations_reportConfigs.Create(
        create_request)

  def _get_filters_for_list(self, source_bucket, destination):
    """Returns the filter string used for list API call."""
    filter_list = []
    if source_bucket is not None:
      filter_list.append(
          'objectMetadataReportOptions.storageFilters.bucket="{}"'.format(
              source_bucket.bucket_name))
    # TODO(b/255962994): Not used currently. Will be tested when we bring
    # back destination support.
    if destination is not None:
      filter_list.append(
          'objectMetadataReportOptions.storageDestinationOptions.'
          'bucket="{}"'.format(destination.bucket_name))
      if destination.resource_name is not None:
        filter_list.append(
            'objectMetadataReportOptions.storageDestinationOptions.'
            'destinationPath="{}"'.format(destination.resource_name))
    if filter_list:
      return ' AND '.join(filter_list)
    else:
      return None

  def list_inventory_reports(
      self, source_bucket=None, destination=None, location=None, page_size=None
  ):
    """Lists the report configs.

    Args:
      source_bucket (storage_url.CloudUrl): Source bucket for which reports will
        be generated.
      destination (storage_url.CloudUrl): The destination url where the
        generated reports will be stored.
      location (str): The location for which the report configs should be
        listed.
      page_size (int|None): Number of items per request to be returend.

    Returns:
      List of Report configs.
    """
    if location is not None:
      parent = _get_parent_string(properties.VALUES.core.project.Get(),
                                  location)
    else:
      parent = _get_parent_string_from_bucket(
          source_bucket.bucket_name
          if source_bucket is not None else destination.bucket_name)
    return list_pager.YieldFromList(
        self.client.projects_locations_reportConfigs,
        self.messages.StorageinsightsProjectsLocationsReportConfigsListRequest(
            parent=parent,
            filter=self._get_filters_for_list(source_bucket, destination)),
        batch_size=page_size if page_size is not None else PAGE_SIZE,
        batch_size_attribute='pageSize',
        field='reportConfigs')

  def get_inventory_report(self, report_config_name):
    """Gets the report config."""
    return self.client.projects_locations_reportConfigs.Get(
        self.messages.StorageinsightsProjectsLocationsReportConfigsGetRequest(
            name=report_config_name))

  def delete_inventory_report(self, report_config_name, force=False):
    """Deletes the report config."""
    request = (
        self.messages
        .StorageinsightsProjectsLocationsReportConfigsDeleteRequest(
            name=report_config_name, force=force))
    return self.client.projects_locations_reportConfigs.Delete(request)

  def _get_frequency_options_and_update_mask(self, start_date, end_date,
                                             frequency):
    """Returns a tuple of messages.FrequencyOptions and update_mask list."""
    update_mask = []
    if start_date is not None:
      start_date_message = self.messages.Date(
          year=start_date.year, month=start_date.month, day=start_date.day)
      update_mask.append('frequencyOptions.startDate')
    else:
      start_date_message = None
    if end_date is not None:
      end_date_message = self.messages.Date(
          year=end_date.year, month=end_date.month, day=end_date.day)
      update_mask.append('frequencyOptions.endDate')
    else:
      end_date_message = None
    if frequency is not None:
      frequency_message = getattr(
          self.messages.FrequencyOptions.FrequencyValueValuesEnum,
          frequency.upper())
      update_mask.append('frequencyOptions.frequency')
    else:
      frequency_message = None
    return (
        self.messages.FrequencyOptions(
            startDate=start_date_message,
            endDate=end_date_message,
            frequency=frequency_message),
        update_mask)

  def _get_metadata_options_and_update_mask(self, metadata_fields,
                                            destination_url):
    """Returns a tuple of messages.ObjectMetadataReportOptions and update_mask."""
    update_mask = []
    if metadata_fields:
      update_mask.append('objectMetadataReportOptions.metadataFields')
    if destination_url is not None:
      storage_destination_message = (
          self.messages.CloudStorageDestinationOptions(
              bucket=destination_url.bucket_name,
              destinationPath=destination_url.resource_name,
          )
      )
      update_mask.append(
          'objectMetadataReportOptions.storageDestinationOptions.bucket')
      update_mask.append(
          'objectMetadataReportOptions.storageDestinationOptions'
          '.destinationPath')
    else:
      storage_destination_message = None
    return (self.messages.ObjectMetadataReportOptions(
        metadataFields=metadata_fields,
        storageDestinationOptions=storage_destination_message), update_mask)

  def _get_report_format_options_and_update_mask(
      self, csv_separator, csv_delimiter, csv_header, parquet):
    """Returns a tuple of ReportFormatOptions and update_mask list."""
    report_format_options = self._get_report_format_options(
        csv_separator, csv_delimiter, csv_header, parquet)
    update_mask = []
    if report_format_options.parquet is not None:
      update_mask.append('parquetOptions')
    else:
      if csv_delimiter is not None:
        update_mask.append('csvOptions.delimiter')
      if csv_header is not None:
        update_mask.append('csvOptions.headerRequired')
      if csv_separator is not None:
        update_mask.append('csvOptions.recordSeparator')
    return (report_format_options, update_mask)

  def update_inventory_report(
      self,
      report_config_name,
      destination_url=None,
      metadata_fields=None,
      start_date=None,
      end_date=None,
      frequency=None,
      csv_separator=None,
      csv_delimiter=None,
      csv_header=None,
      parquet=None,
      display_name=None,
  ):
    """Updates a report config.

    Args:
      report_config_name (str): The name of the report config to be modified.
      destination_url (storage_url.CloudUrl): The destination url where the
        generated reports will be stored.
      metadata_fields (list[str]): Fields to be included in the report.
      start_date (datetime.datetime.date): The date to start generating reports.
      end_date (datetime.datetime.date): The date after which to stop generating
        reports.
      frequency (str): Can be either DAILY or WEEKLY.
      csv_separator (str): The character used to separate the records in the
        CSV file.
      csv_delimiter (str): The delimiter that separates the fields in the CSV
        file.
      csv_header (bool): If True, include the headers in the CSV file.
      parquet (bool): If True, set the parquet options.
      display_name (str): Display name for the report config.

    Returns:
      The created ReportConfig object.
    """
    frequency_options, frequency_update_mask = (
        self._get_frequency_options_and_update_mask(
            start_date, end_date, frequency))
    object_metadata_report_options, metadata_update_mask = (
        self._get_metadata_options_and_update_mask(
            metadata_fields, destination_url))

    report_format_options, report_format_mask = (
        self._get_report_format_options_and_update_mask(
            csv_separator, csv_delimiter, csv_header, parquet))

    # Only the fields present in the mask will be updated.
    update_mask = (
        frequency_update_mask + metadata_update_mask + report_format_mask)

    if display_name is not None:
      update_mask.append('displayName')

    if not update_mask:
      raise errors.CloudApiError(
          'Nothing to update for report config: {}'.format(report_config_name))

    report_config = self.messages.ReportConfig(
        csvOptions=report_format_options.csv,
        parquetOptions=report_format_options.parquet,
        displayName=display_name,
        frequencyOptions=frequency_options,
        objectMetadataReportOptions=object_metadata_report_options)
    request = (
        self.messages.StorageinsightsProjectsLocationsReportConfigsPatchRequest(
            name=report_config_name,
            reportConfig=report_config,
            updateMask=','.join(update_mask),
        )
    )
    return self.client.projects_locations_reportConfigs.Patch(request)

  def list_report_details(self, report_config_name, page_size=None):
    """Lists the report details."""
    return list_pager.YieldFromList(
        self.client.projects_locations_reportConfigs_reportDetails,
        self.messages
        .StorageinsightsProjectsLocationsReportConfigsReportDetailsListRequest(
            parent=report_config_name),
        batch_size=page_size if page_size is not None else PAGE_SIZE,
        batch_size_attribute='pageSize',
        field='reportDetails')

  def get_report_details(self, report_config_name):
    return self.client.projects_locations_reportConfigs_reportDetails.Get(
        self.messages
        .StorageinsightsProjectsLocationsReportConfigsReportDetailsGetRequest(
            name=report_config_name))

  def cancel_operation(self, operation_name):
    self.client.projects_locations_operations.Cancel(
        self.messages.StorageinsightsProjectsLocationsOperationsCancelRequest(
            name=operation_name
        )
    )

  def get_operation(self, operation_name):
    return self.client.projects_locations_operations.Get(
        self.messages.StorageinsightsProjectsLocationsOperationsGetRequest(
            name=operation_name
        )
    )

  def list_operations(self, parent_resource_name):
    request = (
        self.messages.StorageinsightsProjectsLocationsOperationsListRequest(
            name=parent_resource_name,
        )
    )
    return list_pager.YieldFromList(
        self.client.projects_locations_operations,
        request,
        batch_size_attribute='pageSize',
        field='operations',
    )



# -*- coding: utf-8 -*- #
# Copyright 2025 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module containing the extended wrappers for Remediation Intents service."""


from typing import Any

from googlecloudsdk.api_lib.scc.remediation_intents import sps_api
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.scc.remediation_intents import errors


class ExtendedSPSClient():
  """Extended client for the SPS Service (wrappers for specific API calls).

  Attributes:
    release_track: The Gcloud release track to use, like ALPHA, GA.
    org_id: The organization ID for which the API methods are called.
    api_version: The API version to use like v1alpha, main etc.
    client: The client for the SPS Service.
    messages: The messages module for the SPS Service.
  """

  def __init__(self, org_id: str, release_track=base.ReleaseTrack.ALPHA):
    """Initializes the ExtendedSPSClient.

    Args:
      org_id: The organization ID for which the API methods are called.
      release_track: The release track to use for the API version.
    """
    self.release_track = release_track
    self.org_id = org_id
    self.api_version = sps_api.VERSION_MAP.get(release_track)
    self.client = sps_api.GetClientInstance(release_track)
    self.messages = sps_api.GetMessagesModule(release_track)

  def fetch_enqueued_remediation_intent(self) -> Any:
    """Fetches a Remediation Intent resource in ENQUEUED state in given org.

    Returns:
      A Remediation Intent resource in ENQUEUED state for the given org. If no
      such resource is found, returns None.
      Return format is of class (securityposture.messages.RemediationIntent).

    Raises:
      APICallError: An error while calling the SPS Service.
    """
    request = self.messages.SecuritypostureOrganizationsLocationsRemediationIntentsListRequest(
        parent=f'organizations/{self.org_id}/locations/global',
        filter='state : REMEDIATION_INTENT_ENQUEUED',
    )
    try:
      response = (  # List API call.
          self.client.organizations_locations_remediationIntents.List(request)
      )
    except Exception as e:  # Any error like network or system failure.
      raise errors.APICallError('List', str(e))

    remediation_intents = response.remediationIntents
    if remediation_intents is None or len(remediation_intents) < 1:
      return None
    return remediation_intents[0]

  def create_semi_autonomous_remediation_intent(self) -> None:
    """Creates a Semi Autonomous type Remediation Intent resource.

    Raises:
      APICallError: An error while calling the SPS Service.
    """
    request = self.messages.SecuritypostureOrganizationsLocationsRemediationIntentsCreateRequest(
        parent=f'organizations/{self.org_id}/locations/global',
        createRemediationIntentRequest=self.messages.CreateRemediationIntentRequest(
            workflowType=self.messages.CreateRemediationIntentRequest.WorkflowTypeValueValuesEnum.WORKFLOW_TYPE_SEMI_AUTONOMOUS,
        ),
    )
    try:  # Create API call.
      operation = self.client.organizations_locations_remediationIntents.Create(
          request=request
      )
      _ = sps_api.WaitForOperation(   # Polling the LRO.
          operation_ref=sps_api.GetOperationsRef(operation.name),
          message='Waiting for remediation intent to be created',
          has_result=True,
      )
    except Exception as e:
      raise errors.APICallError('Create', str(e))

  def update_remediation_intent(
      self,
      ri_name: str, update_mask: str,
      remediation_intent: Any,
  ) -> Any:
    """Updates a Remediation Intent resource.

    Args:
      ri_name: The name of the Remediation Intent resource to be updated.
      update_mask: The update mask for the update operation.
      remediation_intent: The updated Remediation Intent resource.

    Returns:
      The updated Remediation Intent resource.
      Return format is of class (securityposture.messages.RemediationIntent).

    Raises:
      APICallError: An error while calling the SPS Service.
    """
    request = self.messages.SecuritypostureOrganizationsLocationsRemediationIntentsPatchRequest(
        name=ri_name,
        updateMask=update_mask,
        remediationIntent=remediation_intent
    )
    try:
      operation = self.client.organizations_locations_remediationIntents.Patch(
          request=request
      )
      return sps_api.WaitForOperation(   # Polling the LRO.
          operation_ref=sps_api.GetOperationsRef(operation.name),
          message='Waiting for remediation intent to be updated',
          has_result=True,
      )
    except Exception as e:
      raise errors.APICallError('Update', str(e))



# -*- coding: utf-8 -*- #
# Copyright 2014 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command for deleting images."""


from googlecloudsdk.api_lib.compute import base_classes
from googlecloudsdk.api_lib.compute import utils
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.compute import flags as compute_flags
from googlecloudsdk.command_lib.compute.images import flags


class Delete(base.DeleteCommand):
  """Delete Compute Engine images."""

  @staticmethod
  def Args(parser):
    Delete.DiskImageArg = flags.MakeDiskImageArg(plural=True)
    Delete.DiskImageArg.AddArgument(parser, operation_type='delete')
    parser.display_info.AddCacheUpdater(flags.ImagesCompleter)

  def Run(self, args):
    holder = base_classes.ComputeApiHolder(self.ReleaseTrack())
    client = holder.client

    image_refs = Delete.DiskImageArg.ResolveAsResource(
        args,
        holder.resources,
        scope_lister=compute_flags.GetDefaultScopeLister(client))

    utils.PromptForDeletion(image_refs)

    requests = []
    for image_ref in image_refs:
      requests.append((client.apitools_client.images, 'Delete',
                       client.messages.ComputeImagesDeleteRequest(
                           **image_ref.AsDict())))

    return client.MakeRequests(requests)


Delete.detailed_help = {
    'DESCRIPTION':
        '*{command}* deletes one or more Compute Engine images.',
    'EXAMPLES':
        """
        To delete images 'my-image1' and 'my-image2', run:

            $ {command} my-image1 my-image2
        """,
}



# -*- coding: utf-8 -*- #
# Copyright 2025 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Command for adding endpoints to a wire group."""


from googlecloudsdk.api_lib.compute import base_classes
from googlecloudsdk.api_lib.compute.interconnects.wire_groups import client
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.compute import scope as compute_scope
from googlecloudsdk.command_lib.compute.interconnects.cross_site_networks import flags as cross_site_network_flags
from googlecloudsdk.command_lib.compute.interconnects.wire_groups import flags
from googlecloudsdk.core import properties

_DETAILED_HELP = {
    'DESCRIPTION': """\
        *{command}* is used to add endpoints to a wire group.

        For an example, refer to the *EXAMPLES* section below.
        """,
    'EXAMPLES': """\
        To add an endpoint to a wire group, run:

          $ {command} example-wg \
              --cross-site-network=example-csn \
              --endpoint-label=endpoint-1
        """,
}


@base.UniverseCompatible
@base.ReleaseTracks(
    base.ReleaseTrack.ALPHA,
    base.ReleaseTrack.BETA,
    base.ReleaseTrack.GA,
    base.ReleaseTrack.PREVIEW,
)
class AddEndpoint(base.UpdateCommand):
  """Add endpoint to a Compute Engine wire group.

  *{command}* adds endpoint to a Compute Engine wire group.
  """

  # Framework override.
  detailed_help = _DETAILED_HELP

  WIRE_GROUP_ARG = None
  CROSS_SITE_NETWORK_ARG = None

  @classmethod
  def Args(cls, parser):
    cls.CROSS_SITE_NETWORK_ARG = (
        cross_site_network_flags.CrossSiteNetworkArgumentForOtherResource()
    )
    cls.CROSS_SITE_NETWORK_ARG.AddArgument(parser)
    cls.WIRE_GROUP_ARG = flags.WireGroupArgument(plural=False)
    cls.WIRE_GROUP_ARG.AddArgument(parser, operation_type='update')
    flags.AddEndpointLabel(parser)

  def Collection(self):
    return 'compute.wireGroups'

  def Run(self, args):
    holder = base_classes.ComputeApiHolder(self.ReleaseTrack())
    ref = self.WIRE_GROUP_ARG.ResolveAsResource(
        args,
        holder.resources,
        default_scope=compute_scope.ScopeEnum.GLOBAL,
        additional_params={'crossSiteNetwork': args.cross_site_network},
    )
    project = properties.VALUES.core.project.GetOrFail()

    self._messages = holder.client.messages

    wire_group = client.WireGroup(
        ref=ref,
        project=project,
        cross_site_network=args.cross_site_network,
        compute_client=holder.client,
        resources=holder.resources,
    )
    endpoint_label = args.endpoint_label
    endpoints = wire_group.Describe().endpoints

    endpoints_map = convert_endpoints_to_dict(endpoints)

    endpoints_map[endpoint_label] = holder.client.messages.WireGroupEndpoint()

    endpoints = _build_endpoint_messages(self._messages, endpoints_map)

    return wire_group.Patch(
        endpoints=endpoints,
    )


def convert_endpoints_to_dict(endpoints):
  """Extracts the key,value pairs from the additionalProperties attribute.

  Creates a python dict to be able to pass them into the client.

  Args:
    endpoints: the list of additionalProperties messages

  Returns:
    Python dictionary containing the key value pairs.
  """
  endpoints_map = {}

  if not endpoints or not endpoints.additionalProperties:
    return endpoints_map

  for endpoint_property in endpoints.additionalProperties:
    key, value = endpoint_property.key, endpoint_property.value
    endpoints_map[key] = value

  return endpoints_map


def _build_endpoint_messages(messages, endpoints_map):
  """Builds a WireGroup.EndpointValue message.

  This is so we can re-assign them to the additionalProperties attribute on
  the WireGroup.EndpointsValue message.

  Args:
    messages: the messages module
    endpoints_map: map of endpoints with label as the key and the
      endpoint message as the value

  Returns:
    WireGroup.EndpointsValue message
  """
  endpoint_properties_list = []

  for endpoint_label, endpoints_message in endpoints_map.items():
    endpoint_properties_list.append(
        messages.WireGroup.EndpointsValue.AdditionalProperty(
            key=endpoint_label,
            value=endpoints_message,
        )
    )

  return messages.WireGroup.EndpointsValue(
      additionalProperties=endpoint_properties_list
  )



# -*- coding: utf-8 -*- #
# Copyright 2025 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for Cloud Monitoring Alerts API."""


from googlecloudsdk.api_lib.monitoring import util


class AlertsClient(object):
  """Client for the Alert service in the Stackdriver Monitoring API."""

  def __init__(self, client=None, messages=None):
    self.client = client or util.GetClientInstance()
    self.messages = messages or self.client.MESSAGES_MODULE
    self._service = self.client.projects_alerts

  def Get(self, alert_ref):
    """Gets a Monitoring Alert."""
    request = self.messages.MonitoringProjectsAlertsGetRequest(
        name=alert_ref.RelativeName()
    )
    return self._service.Get(request)

  def List(self, project_ref, a_filter=None, order_by=None, page_size=None):
    """Lists Monitoring Alerts."""
    request = self.messages.MonitoringProjectsAlertsListRequest(
        parent=project_ref.RelativeName(),
        filter=a_filter,
        orderBy=order_by,
        pageSize=page_size,
    )
    return self._service.List(request)



# -*- coding: utf-8 -*- #
# Copyright 2026 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Task for moving an object in a cloud."""

import os
import threading
from googlecloudsdk.api_lib.storage import api_factory
from googlecloudsdk.api_lib.storage import request_config_factory
from googlecloudsdk.command_lib.storage import errors
from googlecloudsdk.command_lib.storage import manifest_util
from googlecloudsdk.command_lib.storage import progress_callbacks
from googlecloudsdk.command_lib.storage import storage_url
from googlecloudsdk.command_lib.storage.tasks import task
from googlecloudsdk.command_lib.storage.tasks import task_status
from googlecloudsdk.command_lib.storage.tasks.cp import copy_util
from googlecloudsdk.core import log


class MoveObjectTask(copy_util.ObjectCopyTaskWithExitHandler):
  """A task for moving an object in a cloud.

  Attributes:
    parallel_processing_key: The parallel processing key for the task.
  """

  def __init__(
      self,
      source_resource,
      destination_resource,
      api_client=None,
      print_created_message=False,
      request_config_factory_fn=None,
      user_request_args=None,
      verbose=False,
  ):
    """Initializes the MoveObjectTask.

    Args:
      source_resource (resource_reference.Resource): Must contain the full
        object path.
      destination_resource (resource_reference.Resource): Must contain the full
        object path.
      api_client (cloud_api.CloudApi|None): API client for the cloud.
      print_created_message (bool): See parent class.
      request_config_factory_fn (function|None): Function for generating
        RequestConfig.
      user_request_args (UserRequestArgs|None): See parent class
      verbose (bool): See parent class.
    """
    super(MoveObjectTask, self).__init__(
        source_resource,
        destination_resource,
        print_created_message=print_created_message,
        user_request_args=user_request_args,
        verbose=verbose,
    )

    if (
        source_resource.storage_url.scheme
        != destination_resource.storage_url.scheme
    ) or not isinstance(source_resource.storage_url, storage_url.CloudUrl):
      raise errors.InvalidUrlError(
          'MoveObjectTask takes two URLs from the same cloud provider.'
      )

    self._api_client = api_client
    self._request_config_factory_fn = (
        request_config_factory_fn or request_config_factory.get_request_config
    )
    self.parallel_processing_key = (
        self._destination_resource.storage_url.url_string
    )

  def execute(self, task_status_queue=None):
    api_client = self._api_client or api_factory.get_api(
        self._source_resource.storage_url.scheme
    )

    if copy_util.check_for_cloud_clobber(
        self._user_request_args, api_client, self._destination_resource
    ):
      no_clobber_message = copy_util.get_no_clobber_message(
          self._destination_resource.storage_url
      )
      log.status.Print(no_clobber_message)
      if self._send_manifest_messages:
        manifest_util.send_skip_message(
            task_status_queue,
            self._source_resource,
            self._destination_resource,
            no_clobber_message,
        )
      return None

    progress_callback = progress_callbacks.FilesAndBytesProgressCallback(
        status_queue=task_status_queue,
        offset=0,
        length=self._source_resource.size,
        source_url=self._source_resource.storage_url,
        destination_url=self._destination_resource.storage_url,
        operation_name=task_status.OperationName.INTRA_CLOUD_COPYING,
        process_id=os.getpid(),
        thread_id=threading.get_ident(),
    )

    request_config = self._request_config_factory_fn(
        url=self._destination_resource.storage_url,
        user_request_args=self._user_request_args,
    )

    result_resource = api_client.move_object(
        self._source_resource.storage_url,
        destination_url=self._destination_resource.storage_url,
        request_config=request_config,
        progress_callback=progress_callback,
    )

    if task_status_queue:
      progress_callback(result_resource.size)

    self._print_created_message_if_requested(result_resource)
    if self._send_manifest_messages:
      manifest_util.send_success_message(
          task_status_queue,
          self._source_resource,
          self._destination_resource,
          md5_hash=result_resource.md5_hash,
      )

    return task.Output(additional_task_iterators=None, messages=None)

  def __eq__(self, other):
    if not isinstance(other, MoveObjectTask):
      return NotImplemented
    return (
        self._source_resource == other._source_resource
        and self._destination_resource == other._destination_resource
        and self._api_client == other._api_client
        and self._print_created_message == other._print_created_message
        and self._request_config_factory_fn == other._request_config_factory_fn
        and self._user_request_args == other._user_request_args
        and self._verbose == other._verbose
    )
